from typing import Any, Union, Type, TYPE_CHECKING, List, Set, Tuple, Dict

from mewpy.util.history import HistoryManager, recorder
from mewpy.util.serialization import serialize, Serializer

# Preventing circular dependencies that only happen due to type checking
if TYPE_CHECKING:
    from mewpy.model import Model, MetabolicModel, RegulatoryModel
    from mewpy.mew.algebra import Expression, Symbolic

    from .gene import Gene
    from .interaction import Interaction
    from .metabolite import Metabolite
    from .reaction import Reaction
    from .regulator import Regulator
    from .target import Target


class MetaVariable(type):
    factories = {}

    def __new__(mcs, name, bases, attrs, **kwargs):

        # if it is the variable factory, only registration is done
        factory = kwargs.get('factory', False)

        if factory:
            cls = super(MetaVariable, mcs).__new__(mcs, name, bases, attrs)

            MetaVariable.factories[name] = cls

            return cls

        # Dynamic typing being used. In this case, a proper name and variable type must be provided
        dynamic = kwargs.get('dynamic', False)
        if dynamic:
            names = [base.variable_type for base in bases]

            name = ''.join([name.title() for name in names])
            name += 'Variable'

            kwargs['variable_type'] = '-'.join(names)

        # The variable type is always added to the subclasses. If it is not given upon subclass creation,
        # the subclass name is to be used
        variable_type = kwargs.get('variable_type', name.lower())
        attrs['variable_type'] = variable_type

        return super(MetaVariable, mcs).__new__(mcs, name, bases, attrs)

    def __init__(cls, name, bases, attrs, **kwargs):

        super().__init__(name, bases, attrs)

        factory = kwargs.get('factory', False)
        if factory:
            # collection of all attributes that must be serialized for a particular class or subclass
            attributes = cls.get_serializable_attributes(attrs)

            attrs['_attributes_registry']['variable'] = attributes

            # Skip further building of the Variable factory
            return

        # Dynamic typing being used. In this case, all children have already been constructed, so everything can be
        # skipped
        dynamic = kwargs.get('dynamic', False)
        if dynamic:
            return

        # Several attributes and methods must be added automatically to the Variable factory children based on the
        # variable type.
        variable_type = attrs['variable_type']

        for base in bases:

            factory = MetaVariable.factories.get(base.__name__)

            if factory and hasattr(cls, 'variable_type'):

                # if some class inherits from the Variable factory, it must be registered in the factory for type
                # checking
                if cls.variable_type not in factory.get_registry():
                    raise TypeError(f'{cls.variable_type} does not inherit from Variable')

                # If set otherwise upon class creation, all subclasses of the Variable factory will contribute to
                # their parent factory with an alternative initializer. The polymorphic constructor, e.g. from_{
                # variable_type}, is added to the Variable factory
                constructor = kwargs.get('constructor', True)

                if constructor:
                    # noinspection PyProtectedMember
                    var_type_initializer = Variable._from(cls)
                    setattr(factory, f'from_{variable_type}', var_type_initializer)

                # If set otherwise upon class creation, all subclasses of the Variable factory will contribute to
                # their parent factory with a declared type checker. The type checker, e.g. is_{variable_type},
                # is added to the Variable factory
                checker = kwargs.get('checker', True)

                if checker:
                    # noinspection PyProtectedMember
                    var_type_checker = Variable._is(variable_type)
                    setattr(factory, f'is_{variable_type}', var_type_checker)

                # collection of all attributes that must be serialized for a particular class or sub class
                attributes = cls.get_serializable_attributes(attrs)

                Variable.register_attributes(attributes, cls)

    @staticmethod
    def get_serializable_attributes(attrs):

        # property getter is marked with serialize and deserialize attribute by the serialize decorator. If
        # deserialization is not to be used, deserialize should be marked with None

        attributes = {}

        for name, method in attrs.items():

            if hasattr(method, 'fget'):

                if hasattr(method.fget, 'serialize') and hasattr(method.fget, 'deserialize') and hasattr(method.fget,
                                                                                                         'pickle'):
                    attributes[name] = (method.fget.serialize, method.fget.deserialize, method.fget.pickle)

        return attributes


# TODO: methods stubs
class Variable(Serializer, metaclass=MetaVariable, factory=True):
    # -----------------------------------------------------------------------------
    # Factory management
    # -----------------------------------------------------------------------------

    _registry = {}
    _attributes_registry = {'variable': {}}

    def __init_subclass__(cls, **kwargs):

        super(Variable, cls).__init_subclass__(**kwargs)

        # the child type
        variable_type = getattr(cls, 'variable_type', cls.__name__.lower())

        cls.register_type(variable_type, cls)

    @staticmethod
    def get_registry():
        return Variable._registry.copy()

    @staticmethod
    def get_attributes_registry():
        return Variable._attributes_registry.copy()

    @staticmethod
    def register_type(variable_type, child):

        Variable._registry[variable_type] = child

    # -----------------------------------------------------------------------------
    # Serialization
    # -----------------------------------------------------------------------------

    @staticmethod
    def register_attributes(attributes, child):

        if hasattr(child, 'variable_type'):
            Variable._attributes_registry[child.variable_type] = attributes

        elif child is Variable:
            Variable._attributes_registry['variable'] = attributes

    @property
    def attributes(self):

        class_attributes = self.get_attributes_registry()

        attributes = class_attributes['variable'].copy()

        for variable_type in self.types:

            for name, (serialize_name, deserialize_name, pickle_name) in class_attributes[variable_type].items():
                attributes[name] = (serialize_name, deserialize_name, pickle_name)

        return attributes

    # -----------------------------------------------------------------------------
    # Factory polymorphic constructor
    # -----------------------------------------------------------------------------

    @classmethod
    def factory(cls, *args: str) -> Union[Type['Variable'],
                                          Type['Gene'],
                                          Type['Interaction'],
                                          Type['Metabolite'],
                                          Type['Reaction'],
                                          Type['Regulator'],
                                          Type['Target']]:

        if not args:
            args = ()

        registry = cls.get_registry()

        types = tuple([registry[name] for name in args])

        if len(types) == 1:
            return types[0]

        _Variable = MetaVariable('Variable', types, {}, dynamic=True)

        return _Variable

    # -----------------------------------------------------------------------------
    # Factory polymorphic initializer
    # -----------------------------------------------------------------------------

    @classmethod
    def from_types(cls, types: List[str], **kwargs) -> Union['Variable',
                                                             'Gene',
                                                             'Interaction',
                                                             'Metabolite',
                                                             'Reaction',
                                                             'Regulator',
                                                             'Target']:
        VariableType = cls.factory(*types)
        return VariableType(**kwargs)

    # -----------------------------------------------------------------------------
    # Factory helper methods
    # -----------------------------------------------------------------------------

    @staticmethod
    def _from(cls):

        def from_(identifier, **kwargs):
            return cls(identifier, **kwargs)

        return from_

    @staticmethod
    def _is(variable_type):

        def is_(self):
            return self.is_a(variable_type)

        return is_

    # -----------------------------------------------------------------------------
    # Base initializer
    # -----------------------------------------------------------------------------

    def __init__(self,
                 identifier: Any,
                 name: str = None,
                 aliases: Set[str] = None,
                 model: Union['Model', 'MetabolicModel', 'RegulatoryModel'] = None):

        """

        Variable: The variable is the most base type of a given variable type,
        such as gene, interaction, metabolite, reaction, regulator and target, among others.
        See also variables.Gene, variables.Interaction, variables.Metabolite, variables.Reaction,
        variables.Regulator and variables.Target for concrete implementations of a variable type.

        The variable object is actually the factory for several variable types.
        This variable object provides all batteries to create new variable types
        and to manage the variables types' inheritance.

        The factory type assists with systematic tasks such as attributes and containers serialization,
        history management, among others.

        For that, attributes, containers and inheritance are registered in the factory. The MetaVariable manages
        both the factory and the variable types derived from the factory.

        :param identifier: identifier, e.g. b0001
        :param name: the name of the model
        :param aliases: aliases for this variable
        :param model: the model to which this variable belongs. Since the default is None, variables do not need to
        be directly associated with a model
        """

        self._check_inheritance()

        if not aliases:
            aliases = set()

        if not model:
            model = None

        if not name:
            name = ''

        self._id = identifier
        self._aliases = aliases
        self._model = model
        self._name = name

        self._contexts = []
        self._history = HistoryManager()

        self._model_ref = 0

    def _check_inheritance(self):

        registry = self.get_registry()

        for variable_type in self.types:
            if variable_type not in registry:
                raise ValueError(f'{variable_type} is not registered as subclass of {self.__class__.__name__}')

    # -----------------------------------------------------------------------------
    # Built-in
    # -----------------------------------------------------------------------------

    def __str__(self):
        return self.name

    def __repr__(self):
        if self.types:
            types = [v_type.title() for v_type in self.types]

            return ', '.join(types) + f': {self.id}'

        return f'Variable: {self.id}'

    # -----------------------------------------------------------------------------
    # Variable type manager
    # -----------------------------------------------------------------------------

    @serialize('types', None)
    @property
    def types(self) -> Set[str]:
        return set()

    # -----------------------------------------------------------------------------
    # Static attributes
    # -----------------------------------------------------------------------------

    @serialize('id', None, '_id')
    @property
    def id(self) -> Any:
        return self._id

    @serialize('name', 'name', '_name')
    @property
    def name(self) -> str:
        return self._name

    @serialize('aliases', 'aliases', '_aliases')
    @property
    def aliases(self) -> Set[str]:
        return self._aliases

    @property
    def model(self) -> Union['Model', 'MetabolicModel', 'RegulatoryModel']:
        return self._model

    # -----------------------------------------------------------------------------
    # Static attributes setters
    # -----------------------------------------------------------------------------

    @name.setter
    @recorder
    def name(self, value: str):

        if not value:
            value = ''

        self._name = value

    @aliases.setter
    @recorder
    def aliases(self, value: Set[str]):

        if not value:
            value = set()

        self._aliases = value

    @model.setter
    def model(self, value: Union['Model', 'MetabolicModel', 'RegulatoryModel']):

        if not value:
            value = None

        self._model = value

    # -----------------------------------------------------------------------------
    # History manager
    # -----------------------------------------------------------------------------

    # The history manager follows the command pattern
    @property
    def contexts(self) -> List[HistoryManager]:

        if self.model:
            return self.model.contexts

        return self._contexts

    @property
    def history(self) -> HistoryManager:

        if self.contexts:
            return self.contexts[-1]

        if self.model:
            return self.model.history

        return self._history

    def undo(self):

        self.history.undo()

    def redo(self):

        self.history.redo()

    def clean_history(self):

        self._history = HistoryManager()

    def clean_context(self):

        self.contexts[:-1] = HistoryManager()

    def reset(self):

        self.history.reset()

    def restore(self):

        self.history.restore()

    def __enter__(self):

        self.contexts.append(HistoryManager())

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):

        context = self.contexts.pop()
        context.reset()

    # -----------------------------------------------------------------------------
    # Operations/Manipulations
    # -----------------------------------------------------------------------------

    def update(self,
               name: str = None,
               aliases: Set[str] = None,
               model: Union['Model', 'MetabolicModel', 'RegulatoryModel'] = None):

        if name is not None:
            self.name = name

        if aliases is not None:
            self.aliases.update(set(aliases))

        if model is not None:
            self.model = model

    # -----------------------------------------------------------------------------
    # Abstract implementations
    # -----------------------------------------------------------------------------

    # The following polymorphic initializers are just registered here to avoid type checking errors
    @classmethod
    def from_gene(cls,
                  identifier: Any,
                  name: str = None,
                  aliases: set = None,
                  model: 'Model' = None,
                  coefficients: Union[Set[Union[int, float]], List[Union[int, float]], Tuple[Union[int, float]]] = None,
                  active_coefficient: Union[int, float] = None,
                  reactions: Dict[str, 'Reaction'] = None) -> 'Gene':
        ...

    @classmethod
    def from_interaction(cls,
                         identifier: Any,
                         name: str = None,
                         aliases: set = None,
                         model: 'Model' = None,
                         target: 'Target' = None,
                         regulatory_events: Dict[Union[float, int], 'Expression'] = None) -> 'Interaction':
        ...

    @classmethod
    def from_metabolite(cls,
                        identifier: Any,
                        name: str = None,
                        aliases: set = None,
                        model: 'Model' = None,
                        charge: int = None,
                        compartment: str = None,
                        formula: str = None,
                        reactions: Dict[str, 'Reaction'] = None) -> 'Metabolite':
        ...

    @classmethod
    def from_reaction(cls,
                      identifier: Any,
                      name: str = None,
                      aliases: set = None,
                      model: 'Model' = None,
                      bounds: Tuple[Union[float, int], Union[float, int]] = None,
                      stoichiometry: Dict['Metabolite', Union[float, int]] = None,
                      gpr: 'Expression' = None) -> 'Reaction':
        ...

    @classmethod
    def from_regulator(cls,
                       identifier: Any,
                       name: str = None,
                       aliases: set = None,
                       model: 'Model' = None,
                       coefficients: Set[Union[float, int]] = None,
                       active_coefficient: Union[float, int] = None,
                       interactions: Dict[str, 'Interaction'] = None) -> 'Regulator':
        ...

    @classmethod
    def from_target(cls,
                    identifier: Any,
                    name: str = None,
                    aliases: set = None,
                    model: 'Model' = None,
                    coefficients: Set[Union[float, int]] = None,
                    active_coefficient: Union[float, int] = None,
                    interaction: 'Interaction' = None) -> 'Target':
        ...

    # The following type checkers are just registered here to avoid type checking errors
    def is_gene(self) -> bool:

        ...

    def is_interaction(self) -> bool:

        ...

    def is_metabolite(self) -> bool:

        ...

    def is_reaction(self) -> bool:

        ...

    def is_regulator(self) -> bool:

        ...

    def is_target(self) -> bool:

        ...

    # -----------------------------------------------------------------------------
    # Type checker
    # -----------------------------------------------------------------------------

    def is_a(self, variable_type) -> bool:

        if variable_type in self.types:
            return True

        return False


# -----------------------------------------------------------------------------
# Helper methods
# -----------------------------------------------------------------------------
def variables_from_symbolic(symbolic: 'Symbolic',
                            types: Union[Set[str], List[str], Tuple[str]],
                            model: Union['Model', 'MetabolicModel', 'RegulatoryModel'] = None):
    variables = {}

    for symbol in symbolic.atoms(symbols_only=True):

        if symbol.name in variables:
            continue

        if model:

            variable = model.get(symbol.name)

            if variable:

                for variable_type in types:

                    if not variable.is_a(variable_type):
                        raise TypeError(f'{symbol.name} is not a {variable_type} in model {model.id}')

        else:
            variable = Variable.from_types(identifier=symbol.name,
                                           types=types,
                                           model=model)

        variables[variable.id] = variable

    return variables


def build_variable(types, kwargs):
    return Variable.from_types(types, **kwargs)



from typing import Union

from mewpy.model import Model, MetabolicModel, RegulatoryModel
from mewpy.solvers.solution import Solution, Status
from mewpy.solvers.solver import VarType, Solver
from mewpy.mew.lp import Notification, ConstraintContainer, VariableContainer, MetabolicLinearizer, GPRLinearizer
from mewpy.mew.solution import ModelSolution


# TODO: missing documentation and typing
class FBA(MetabolicLinearizer):

    def __init__(self,
                 model: Union[Model, MetabolicModel, RegulatoryModel],
                 solver: Union[str, Solver, None] = None,
                 build: bool = True,
                 attach: bool = False):
        """
        Flux Balance Analysis (FBA) of a metabolic model. Regular implementation of a FBA for a metabolic model.

        For more details consult: https://dx.doi.org/10.1038%2Fnbt.1614

        :param model: a mewpy Model, MetabolicModel, RegulatoryModel or all. The model is used to retrieve
        variables and constraints to the linear problem

        :param solver: A Solver, CplexSolver, GurobiSolver or OptLangSolver instance.
        Alternatively, the name of the solver is also accepted.
        The solver interface will be used to load and solve a linear problem in a given solver.
        If none, a new solver is instantiated. An instantiated solver may be used but it will be overwritten
        if build is true.

        :param build: Whether to build the linear problem upon instantiation. Default: False
        :param attach: Whether to attach the linear problem to the model upon instantiation. Default: False
        """

        super().__init__(model=model,
                         solver=solver,
                         build=build,
                         attach=attach)

    def build(self):

        if self._variables or self._constraints:
            self.clean()

        if self.model.is_metabolic():
            self.metabolite_reaction_lookup(reactions=self.model.yield_reactions(),
                                            metabolites=self.model.yield_metabolites())

            self.constraints_from_gprs(inplace=True)

            self.update()

            self.set_objective(linear={var.id: value for var, value in self.model.objective.items()},
                               minimize=False)

    def notification(self, notification: Notification):

        if notification.content_type == 'coefficients' and notification.action == 'set':

            # if the notification is a coefficient object, the bounds must be processed accordingly
            # by each method implementation. In the case of FBA, the reactions bounds are inferred
            # from the coefficients and set regularly.
            # Nevertheless, if the coefficients of a gene are being set, the gprs associated with this gene
            # must be evaluated and then the resulting constraints must be set to the reactions

            variable = notification.content.id
            lb, ub = notification.content.bounds
            active_coefficient = notification.content.active_coefficient

            if variable in self._cols:

                # the variable is a reaction, and bounds must be set normally

                return super(FBA, self).set_bounds(variable=variable, lb=lb, ub=ub)

            else:

                # the variable is a gene, so new constraints must be inferred
                # from the gene coefficients and gprs evaluation

                return self.constraints_from_gene(variable=variable, lb=active_coefficient, ub=ub, inplace=True)

        else:
            return super(FBA, self).notification(notification)

    def optimize(self,
                 objective=None,
                 minimize=False,
                 constraints=None,
                 to_solver=False,
                 get_values=True,
                 shadow_prices=False,
                 reduced_costs=False,
                 pool_size=0,
                 pool_gap=None) -> Union[ModelSolution, Solution]:

        self._set_objective(linear=objective, minimize=minimize)

        if constraints:
            constraints = constraints.copy()

        solver_solution = self._solver.solve(linear=objective,
                                             quadratic=None,
                                             minimize=minimize,
                                             model=None,
                                             constraints=constraints,
                                             get_values=get_values,
                                             shadow_prices=shadow_prices,
                                             reduced_costs=reduced_costs,
                                             pool_size=pool_size,
                                             pool_gap=pool_gap)

        if to_solver:
            return solver_solution

        if minimize:
            sense = 'minimize'

        else:
            sense = 'maximize'

        return ModelSolution.from_solver(method=self.method, solution=solver_solution, model=self.model,
                                         objective_direction=sense)

    def constraints_from_gprs(self, inplace=False):

        rxns, lbs, ubs = [], [], []

        gene_state = {gene.id: gene.coefficient.active_coefficient for gene in self.model.yield_genes()}

        if self.model.is_metabolic():

            # gprs over reactions
            for rxn in self.model.yield_reactions():

                if rxn.gpr.is_none:
                    continue

                else:

                    res = rxn.gpr(values=gene_state)

                    if not res:

                        variable = rxn.id
                        lb = 0.0
                        ub = 0.0

                        if inplace:
                            super(FBA, self).set_bounds(variable=variable, lb=lb, ub=ub)

                        rxns.append(rxn.id)
                        lbs.append(0.0)
                        ubs.append(0.0)

        return rxns, lbs, ubs

    def constraints_from_gene(self, variable, lb, ub, inplace=False):

        rxns, rxn_lbs, rxn_ubs = [], [], []

        gene_state = {gene.id: gene.coefficient.active_coefficient for gene in self.model.yield_genes()}
        gene_state[variable] = lb

        # Either the gene is not available in the model or it is not a gene. The later can occur if a multi type model
        # has an fba simulator attached but a the bounds of a regulator are changed
        model_var = self.model.get(variable, None)

        if model_var is None:
            return

        if not model_var.is_gene():
            return

        for rxn in model_var.yield_reactions():

            if rxn.gpr.is_none:

                continue

            else:

                res = rxn.gpr(values=gene_state)

                if not res:
                    rxn_lb = 0.0
                    rxn_ub = 0.0

                else:
                    rxn_lb = rxn.lower_bound
                    rxn_ub = rxn.upper_bound

                rxn_id = rxn.id
                rxns.append(rxn_id)
                rxn_lbs.append(rxn_lb)
                rxn_ubs.append(rxn_ub)

                if inplace:
                    super(FBA, self).set_bounds(variable=rxn_id, lb=rxn_lb, ub=rxn_ub)

        return rxns, rxn_lbs, rxn_ubs


class milpFBA(MetabolicLinearizer, GPRLinearizer):

    def __init__(self,
                 model: Union[Model, MetabolicModel, RegulatoryModel],
                 solver: Union[str, Solver, None] = None,
                 build: bool = True,
                 attach: bool = False):
        """
        Mixed-Integer Flux Balance Analysis (FBA) of a metabolic model.
        Regular implementation of a FBA for a metabolic model.
        Additionally, GPRs are linearized into mixed-integer constraints, so that genes are variables of the linear
        problem to be solved.

        Check the mewpy.mew.lp.linearizers module for more detail regarding GPRs linearization

        :param model: a mewpy Model, MetabolicModel, RegulatoryModel or all. The model is used to retrieve
        variables and constraints to the linear problem

        :param solver: A Solver, CplexSolver, GurobiSolver or OptLangSolver instance.
        Alternatively, the name of the solver is also accepted.
        The solver interface will be used to load and solve a linear problem in a given solver.
        If none, a new solver is instantiated. An instantiated solver may be used but it will be overwritten
        if build is true.

        :param build: Whether to build the linear problem upon instantiation. Default: False
        :param attach: Whether to attach the linear problem to the model upon instantiation. Default: False
        """

        super().__init__(model=model,
                         solver=solver,
                         build=build,
                         attach=attach)

    def build(self):

        if self._variables or self._constraints:
            self.clean()

        if self.model.is_metabolic():
            self.metabolite_reaction_lookup(reactions=self.model.yield_reactions(),
                                            metabolites=self.model.yield_metabolites())

            self.add_gprs(self.model.yield_reactions())

            self.update()

            self.set_objective(linear={var.id: value for var, value in self.model.objective.items()},
                               minimize=False)

    def notification(self, notification: Notification):

        if notification.content_type == 'coefficients' and notification.action == 'set':

            # if the notification is a coefficient object, the whole gpr must be build again,
            # as lower and upper bounds of variables are also encoded into the constraints

            self.update_milp_constraints(notification.content.id)

            lb, ub = notification.content.bounds

            return super(milpFBA, self).set_bounds(notification.content.id, lb=lb, ub=ub)

        else:
            return super(milpFBA, self).notification(notification)

    def update_milp_constraints(self, reaction):

        variable = self.model.get(reaction)

        if variable.is_reaction():
            self.add_gprs([variable])

    def _optimize_with_constraints(self,
                                   objective=None,
                                   minimize=False,
                                   constraints=None,
                                   get_values=True,
                                   shadow_prices=False,
                                   reduced_costs=False,
                                   pool_size=0,
                                   pool_gap=None):

        constraints = constraints.copy()

        old_simulators = self.model.simulators.copy()

        self.model._simulators = [self]

        with self.model:

            for variable, bounds in constraints.items():
                variable_obj = self.model.get(variable)

                # sending notification. Notification processing will update variables and constraints accordingly
                if isinstance(bounds, (tuple, list, set)):
                    variable_obj.coefficient.coefficients = bounds

                else:
                    variable_obj.coefficient.coefficients = (bounds,)

            # updating the last modifications
            self.update()

            # solving for the last modifications
            solver_solution = self._solver.solve(linear=objective,
                                                 quadratic=None,
                                                 minimize=minimize,
                                                 model=None,
                                                 constraints=None,
                                                 get_values=get_values,
                                                 shadow_prices=shadow_prices,
                                                 reduced_costs=reduced_costs,
                                                 pool_size=pool_size,
                                                 pool_gap=pool_gap)

        # context exit. restore. update
        self.update()

        self.model._simulators = old_simulators

        return solver_solution

    def optimize(self,
                 objective=None,
                 minimize=False,
                 constraints=None,
                 to_solver=False,
                 get_values=True,
                 shadow_prices=False,
                 reduced_costs=False,
                 pool_size=0,
                 pool_gap=None):

        self._set_objective(linear=objective, minimize=minimize)

        if constraints:
            solver_solution = self._optimize_with_constraints(objective=objective,
                                                              minimize=minimize,
                                                              constraints=constraints,
                                                              get_values=get_values,
                                                              shadow_prices=shadow_prices,
                                                              reduced_costs=reduced_costs,
                                                              pool_size=pool_size,
                                                              pool_gap=pool_gap)

        else:
            solver_solution = self._solver.solve(linear=objective,
                                                 quadratic=None,
                                                 minimize=minimize,
                                                 model=None,
                                                 constraints=None,
                                                 get_values=get_values,
                                                 shadow_prices=shadow_prices,
                                                 reduced_costs=reduced_costs,
                                                 pool_size=pool_size,
                                                 pool_gap=pool_gap)

        if to_solver:
            return solver_solution

        if minimize:
            sense = 'minimize'

        else:
            sense = 'maximize'

        return ModelSolution.from_solver(method=self.method, solution=solver_solution, model=self.model,
                                         objective_direction=sense)


class pFBA(FBA):

    def __init__(self,
                 model: Union[Model, MetabolicModel, RegulatoryModel],
                 solver: Union[str, Solver, None] = None,
                 build: bool = True,
                 attach: bool = False):
        """
        Parsimonious Flux Balance Analysis (FBA) of a metabolic model.
        Regular implementation of a pFBA for a metabolic model.

        This pFBA implementation was heavily inspired by pFBA implementation of reframed python package. Take a look at
        the source: https://github.com/cdanielmachado/reframed and https://reframed.readthedocs.io/en/latest/

        For more details consult: https://doi.org/10.1038/msb.2010.47

        :param model: a mewpy Model, MetabolicModel, RegulatoryModel or all. The model is used to retrieve
        variables and constraints to the linear problem

        :param solver: A Solver, CplexSolver, GurobiSolver or OptLangSolver instance.
        Alternatively, the name of the solver is also accepted.
        The solver interface will be used to load and solve a linear problem in a given solver.
        If none, a new solver is instantiated. An instantiated solver may be used but it will be overwritten
        if build is true.

        :param build: Whether to build the linear problem upon instantiation. Default: False
        :param attach: Whether to attach the linear problem to the model upon instantiation. Default: False
        """

        super().__init__(model=model,
                         solver=solver,
                         build=build,
                         attach=attach)

    def build(self):

        super(pFBA, self).build()

    def optimize(self,
                 objective=None,
                 minimize=False,
                 fraction=None,
                 constraints=None,
                 reactions=None,
                 to_solver=False,
                 get_values=True,
                 shadow_prices=False,
                 reduced_costs=False,
                 pool_size=0,
                 pool_gap=None) -> Union[ModelSolution, Solution]:

        self._set_objective(linear=objective, minimize=minimize)

        if constraints:
            constraints = constraints.copy()

        sol = self._solver.solve(linear=objective,
                                 quadratic=None,
                                 minimize=minimize,
                                 model=None,
                                 constraints=constraints,
                                 get_values=False,
                                 shadow_prices=False,
                                 reduced_costs=False,
                                 pool_size=0,
                                 pool_gap=None)

        if sol.status != Status.OPTIMAL:
            return sol

        constraint = ConstraintContainer(name='pfba_constraints', coefs=[], lbs=[], ubs=[])
        variable = VariableContainer(name='pfba_variables', sub_variables=[], lbs=[], ubs=[], variables_type=[])

        constraint.coefs.append(self._linear_objective)
        if fraction is None:

            constraint.lbs.append(sol.fobj)
            constraint.ubs.append(sol.fobj)

        else:

            constraint.lbs.append(sol.fobj * fraction)
            constraint.ubs.append(sol.fobj)

        if not reactions:
            reactions = self.model.reactions.keys()

        pfba_objective = {}

        for rxn in reactions:

            rxn_obj = self.model.get(rxn)

            if rxn_obj.reversibility:

                rxn_forward = f'{rxn}_forward'
                rxn_reverse = f'{rxn}_reverse'

                variable.sub_variables.extend([rxn_forward, rxn_reverse])
                variable.lbs.extend([0, 0])
                variable.ubs.extend([rxn_obj.upper_bound, rxn_obj.upper_bound])
                variable.variables_type.extend([VarType.CONTINUOUS, VarType.CONTINUOUS])

                constraint.coefs.extend([{rxn: -1, rxn_forward: 1},
                                         {rxn: 1, rxn_reverse: 1}])
                constraint.lbs.extend([0, 0])
                constraint.ubs.extend([rxn_obj.upper_bound, rxn_obj.upper_bound])

                pfba_objective[rxn_forward] = 1
                pfba_objective[rxn_reverse] = 1

            else:

                pfba_objective[rxn_obj.id] = 1

        self.add((variable, constraint))
        self.update()

        old_linear_objective = self._linear_objective.copy()
        old_quadratic_objective = self._quadratic_objective.copy()
        old_sense = self._minimize

        self._set_objective(linear=pfba_objective, minimize=True)

        solver_solution = self._solver.solve(linear=pfba_objective,
                                             quadratic=None,
                                             minimize=True,
                                             model=None,
                                             constraints=constraints,
                                             get_values=get_values,
                                             shadow_prices=shadow_prices,
                                             reduced_costs=reduced_costs,
                                             pool_size=pool_size,
                                             pool_gap=pool_gap)

        self.remove((variable, constraint))
        self.update()
        self.set_objective(linear=old_linear_objective, quadratic=old_quadratic_objective, minimize=old_sense)

        if to_solver:
            return solver_solution

        if minimize:
            sense = 'minimize'

        else:
            sense = 'maximize'

        return ModelSolution.from_solver(method=self.method, solution=solver_solution, model=self.model,
                                         objective_direction=sense)



from typing import Union, List, Tuple, Set, Dict, TYPE_CHECKING

if TYPE_CHECKING:
    from mewpy.mew.variables.coefficient import Coefficient


class Notification:

    def __init__(self,
                 content: Union[Dict, List, Tuple, Set, 'Coefficient'],
                 content_type: str,
                 action: str):

        content_types = ('reactions', 'metabolites', 'gprs', 'interactions', 'coefficients', 'objectives')
        actions = ('add', 'remove', 'set')

        if content_type not in content_types:
            raise ValueError(f'Wrong content_type, select one of the following: {content_types}')

        if action not in actions:
            raise ValueError(f'Wrong action, select one of the following: {actions}')

        self.content = content
        self.content_type = content_type
        self.action = action



import numpy as np

from ..simulation import get_simulator
from ..simulation.simulation import Simulator


def flux_envelope(model, r_x, r_y, steps=10, constraints=None):
    """ Calculate the flux envelope for a pair of reactions.
        Adapted from REFRAMED to be compatible both with REFRAMED and COBRApy.


    :param model : The model or simulator.
    :param str r_x: Reaction on x-axis.
    :param str r_y: Reaction on y-axis.
    :param int steps: Number of steps to compute (default: 10).
    :param dict constraints: Custom constraints to the FBA problem.
    :param dict envcond: Environmental conditions.
    :returns:  x values, y_min values, y_max values

    """

    if isinstance(model, Simulator):
        simul = model
    else:
        try:
            simul = get_simulator(model)
        except Exception:
            raise ValueError(
                'The model should be an instance of model or simulator')

    obj_frac = 0
    # if r_x in simul.get_objective():
    #    obj_frac = 0.0

    x_range = simul.FVA(obj_frac=obj_frac, reactions=[r_x], constraints=constraints)
    xmin, xmax = x_range[r_x]
    xvals = np.linspace(xmin, xmax, steps)
    ymins, ymaxs = np.zeros(steps), np.zeros(steps)

    if constraints is None:
        _constraints = {}
    else:
        _constraints = {}
        _constraints.update(constraints)

    for i, xval in enumerate(xvals):
        _constraints[r_x] = xval
        y_range = simul.FVA(obj_frac=obj_frac, reactions=[r_y], constraints=_constraints)
        ymins[i], ymaxs[i] = y_range[r_y]

    return xvals, ymins, ymaxs


def plot_flux_envelope(model, r_x, r_y, steps=10, substrate=None, constraints=None,
                       label_x=None, label_y=None, flip_x=False, flip_y=False,
                       plot_kwargs=None, fill_kwargs=None, ax=None):
    """ Plots the flux envelope for a pair of reactions.
        Adapted from REFRAMED.

    :param model: The model or simulator.
    :param str r_x: Reaction on x-axis.
    :param str r_y: Reaction on y-axis.
    :param int steps: Number of steps to compute (default: 20).
    :param str substrate: Compute yields for given substrate instead of rates (optional).
    :param dict constraints: Additional simulation constraints.
    :param str label_x: x label (optional, uses reaction name by default).
    :param str label_y: y label (optional, uses reaction name by default).
    :param bool flip_x: Flip direction of r_x (default: False).
    :param dict flip_y: Flip direction of r_y (default: False).
    :param dict plot_kwargs: Additional parameters to *pyplot.plot* (optional).
    :param dict fill_kwargs: Additional parameters to *pyplot.fill_between* (optional).
    :param matplotlib.Axes ax: Plot over existing axes (optional).
    :returns:  matplotlib.Axes: Axes object.
    """

    try:
        import matplotlib.pyplot as plt
    except ImportError:
        raise RuntimeError("Matplotlib is not installed.")

    if isinstance(model, Simulator):
        simul = model
    else:
        try:
            simul = get_simulator(model)
        except Exception:
            raise ValueError(
                'model should be an instance of model or simulator')

    offset = 0.03

    if ax is None:
        _, ax = plt.subplots()

    if not plot_kwargs:
        plot_kwargs = {'color': 'k'}

    if not fill_kwargs:
        fill_kwargs = {'color': 'k', 'alpha': 0.1}

    xvals, ymins, ymaxs = flux_envelope(model, r_x, r_y, steps, constraints)

    if flip_x:
        xvals, ymins, ymaxs = -xvals, ymins[::-1], ymaxs[::-1]

    if flip_y:
        ymins, ymaxs = -ymaxs, -ymins

    if substrate:
        sol = simul.simulate()
        uptk = abs(sol.fluxes[substrate])
        xvals, ymins, ymaxs = xvals / uptk, ymins / uptk, ymaxs / uptk

    ax.plot(xvals, ymins, **plot_kwargs)
    ax.plot(xvals, ymaxs, **plot_kwargs)
    ax.plot([xvals[0], xvals[0]], [ymins[0], ymaxs[0]], **plot_kwargs)
    ax.plot([xvals[-1], xvals[-1]], [ymins[-1], ymaxs[-1]], **plot_kwargs)

    ax.fill_between(xvals, ymins, ymaxs, **fill_kwargs)

    ax.set_xlabel(label_x) if label_x else ax.set_xlabel(r_x)
    ax.set_ylabel(label_y) if label_y else ax.set_ylabel(r_y)

    xmin, xmax = min(xvals), max(xvals)
    dx = offset * (xmax - xmin)
    ax.set_xlim((xmin - dx, xmax + dx))

    ymin, ymax = min(ymins), max(ymaxs)
    dy = offset * (ymax - ymin)
    ax.set_ylim((ymin - dy, ymax + dy))

    return ax



from typing import Union, List, Tuple, TYPE_CHECKING, Set

from mewpy.util.history import recorder
from mewpy.util.constants import ModelConstants
from mewpy.mew.algebra import Symbolic
from mewpy.mew.lp import Notification

if TYPE_CHECKING:
    from .variable import Variable
    from .gene import Gene
    from .interaction import Interaction
    from .metabolite import Metabolite
    from .reaction import Reaction
    from .regulator import Regulator
    from .target import Target


# TODO: methods stubs and type hinting
# TODO: Coefficient can be further integrated with the variables that use this
#  object. For that, coefficient composition pattern may be applied
class Coefficient:

    def __init__(self,
                 variable: 'Variable',
                 coefficients: Union[Set[Union[int, float]], List[Union[int, float]], Tuple[Union[int, float]]] = None,
                 active: Union[int, float] = None):

        """

        A coefficient handles multiple coefficients and bounds for a given variable of any type.
        Coefficient object holds many attributes regularly associated with variables,
        such as bounds, coefficients, active coefficient, among others. These attributes are dynamically generated based
        on the coefficients provided as input.

        A coefficient is particularly useful for variables that can take multiple states

        :param variable: variable to which the coefficient belongs to
        :param coefficients: all coefficients that the variable can take
        :param active: the active coefficient
        """

        # variable is a must
        if not variable:
            raise ValueError('A coefficient strictly requires a variable')

        # boolean behavior if an empty coefficient is built
        if not coefficients:
            coefficients = (0.0, 1.0)

        self._variable = variable

        # all coefficients are recorded under data.
        self._data = list(coefficients)

        # the active coefficient defines the current level/bound that a variable can take in a linear problem.
        # by default it is set to the minimum possible
        if active is None:
            active = min(coefficients)

        else:

            if active not in self._data:
                raise ValueError(f'active coefficient {active} is not available in coefficients')

        self._active_coefficient = active

    # -----------------------------------------------------------------------------
    # Built-in
    # -----------------------------------------------------------------------------

    def __str__(self):
        return str(self.coefficients)

    # -----------------------------------------------------------------------------
    # Polymorphic constructors
    # -----------------------------------------------------------------------------

    @classmethod
    def from_bounds(cls,
                    variable: 'Variable',
                    lb: Union[int, float] = None,
                    ub: Union[int, float] = None):

        # useful for reactions

        if lb is None:
            lb = ModelConstants.REACTION_LOWER_BOUND

        if ub is None:
            ub = ModelConstants.REACTION_UPPER_BOUND

        return cls(variable=variable,
                   coefficients=(lb, ub))

    # -----------------------------------------------------------------------------
    # Dynamic attributes
    # -----------------------------------------------------------------------------

    @property
    def coefficients(self) -> List[Union[int, float]]:

        return self._data.copy()

    @property
    def active_coefficient(self) -> float:
        return float(self._active_coefficient)

    @property
    def is_active(self) -> bool:
        return self.active_coefficient > ModelConstants.TOLERANCE

    @property
    def minimum_coefficient(self) -> Union[int, float]:

        return min(self._data)

    @property
    def maximum_coefficient(self) -> Union[int, float]:

        return max(self._data)

    @property
    def lower_bound(self) -> Union[int, float]:

        return self.minimum_coefficient

    @property
    def upper_bound(self) -> Union[int, float]:

        return self.maximum_coefficient

    @property
    def bounds(self) -> Tuple[Union[int, float], Union[int, float]]:

        return self.minimum_coefficient, self.maximum_coefficient

    # -----------------------------------------------------------------------------
    # Variable-based attributes
    # -----------------------------------------------------------------------------

    @property
    def variable(self) -> Union['Variable', 'Gene', 'Interaction', 'Metabolite', 'Reaction', 'Regulator', 'Target']:
        return self._variable

    @property
    def id(self):
        return self.variable.id

    @property
    def model(self):
        return self.variable.model

    @property
    def history(self):
        return self.variable.history

    # -----------------------------------------------------------------------------
    # Operations/Manipulations
    # -----------------------------------------------------------------------------

    @active_coefficient.setter
    @recorder
    def active_coefficient(self, value: Union[int, float]):

        if value not in self._data:
            raise ValueError(f'{value} is not available for this coefficient object')

        self._active_coefficient = value

        self.notify()

    def _setter(self,
                coefficients: Union[Set[Union[int, float]], List[Union[int, float]], Tuple[Union[int, float]]],
                active=None):

        """
        Internal use only

        :param coefficients: coefficients to be set
        :param active: the active coefficient
        :return:
        """

        self._data = list(coefficients)

        if active is None:

            if self.active_coefficient not in self._data:
                self._active_coefficient = self.minimum_coefficient

        else:

            if active in self._data:
                self._active_coefficient = active

            else:
                self._active_coefficient = self.minimum_coefficient

        self.notify()

    @coefficients.setter
    @recorder
    def coefficients(self, value: Union[Set[Union[int, float]], List[Union[int, float]], Tuple[Union[int, float]]]):

        self._setter(value)

    def __getitem__(self, item: int) -> Union[int, float]:

        return self._data[item]

    def __setitem__(self, key: int, value: Union[int, float], history=True):

        old_coef = self.__getitem__(key)

        self._data[value] = value

        self.notify()

        if history:
            self.history.queue_command(undo_func=self.__setitem__,
                                       undo_kwargs={'key': key,
                                                    'value': old_coef,
                                                    'history': False},
                                       func=self.__setitem__,
                                       kwargs={'key': key,
                                               'value': value,
                                               'history': False})

    def get(self, index: int, default=None) -> Union[int, float]:

        try:
            return self.__getitem__(index)

        except IndexError:

            return default

    def index(self, value: Union[int, float], default=None) -> int:

        try:
            return self._data.index(value)

        except ValueError:

            return default

    def _add(self, coefficient: Union[int, float], active: bool = False):

        """
        Internal use only

        :param coefficient:
        :param active:
        :return:
        """

        self._data.append(coefficient)

        if active:
            self._active_coefficient = coefficient

    def add(self, coefficient: Union[int, float], active: bool = False, history=True):

        if history:
            self.history.queue_command(undo_func=self.remove,
                                       undo_kwargs={'coefficient': coefficient,
                                                    'history': False},
                                       func=self.add,
                                       kwargs={'coefficient': coefficient,
                                               'active': active,
                                               'history': False})

        self._add(coefficient=coefficient, active=active)

        self.notify()

    def remove(self, coefficient: Union[int, float], history=True) -> Union[int, float]:

        removed = self._data.remove(coefficient)

        is_active = False

        if coefficient == self._active_coefficient:
            is_active = True

        if is_active:
            self._active_coefficient = self.minimum_coefficient

        if history:
            self.history.queue_command(undo_func=self.add,
                                       undo_kwargs={'coefficient': removed,
                                                    'active': is_active,
                                                    'history': False},
                                       func=self.remove,
                                       kwargs={'coefficient': coefficient,
                                               'history': False})

        self.notify()

        return removed

    def extend(self,
               coefficients: Union[Set[Union[int, float]], List[Union[int, float]], Tuple[Union[int, float]]],
               active_coefficient: float = None,
               history=True):

        old_coefficients = self.coefficients
        old_active = self.active_coefficient

        for coefficient in coefficients:

            active = False

            if coefficient == active_coefficient:
                active = True

            self._add(coefficient, active=active)

        if history:
            self.history.queue_command(undo_func=self._setter,
                                       undo_kwargs={'coefficients': old_coefficients,
                                                    'active': old_active},
                                       func=self.extend,
                                       kwargs={'coefficients': coefficients,
                                               'history': False})

        self.notify()

    def ko(self, minimum_coefficient: Union[int, float] = None, history=True):

        if not minimum_coefficient:
            minimum_coefficient = self.minimum_coefficient

        old_coefficients = self.coefficients
        old_active = self.active_coefficient

        self._data = [minimum_coefficient]
        self._active_coefficient = minimum_coefficient

        if history:
            self.history.queue_command(undo_func=self._setter,
                                       undo_kwargs={'coefficients': old_coefficients,
                                                    'active': old_active},
                                       func=self.ko,
                                       kwargs={'minimum_coefficient': minimum_coefficient,
                                               'history': False})

        self.notify()

    def activate(self, maximum_coefficient: Union[int, float] = None, history=True):

        if not maximum_coefficient:
            maximum_coefficient = self.maximum_coefficient

        old_coefficients = self.coefficients
        old_active = self.active_coefficient

        self._data = [maximum_coefficient]
        self._active_coefficient = maximum_coefficient

        if history:
            self.history.queue_command(undo_func=self._setter,
                                       undo_kwargs={'coefficients': old_coefficients,
                                                    'active': old_active,
                                                    'history': False},
                                       func=self.activate,
                                       kwargs={'maximum_coefficient': maximum_coefficient,
                                               'history': False})

        self.notify()

    # -----------------------------------------------------------------------------
    # Simulators interface
    # -----------------------------------------------------------------------------

    def notify(self):

        if self.model:

            notification = Notification(content=self,
                                        content_type='coefficients',
                                        action='set')

            self.model.notify(notification)



"""odespy ODE Solver
"""
from .ode import ODEMethod, SolverConfigurations, ODESolver
import odespy

methods = {
    ODEMethod.LSODA: odespy.Lsoda,
    ODEMethod.LSODAR: odespy.Lsodar,
    ODEMethod.LSODE: odespy.Lsode,
    ODEMethod.HEUN: odespy.Heun,
    ODEMethod.EULER: odespy.Euler,
    ODEMethod.RK4: odespy.RK4,
    ODEMethod.DORMAN_PRINCE: odespy.DormandPrince,
    ODEMethod.RKFehlberg: odespy.RKFehlberg,
    ODEMethod.Dopri5: odespy.Dopri5,
    # ODEMethod.Dop853: odespy.Dop853,
    ODEMethod.Vode: odespy.Vode,
    ODEMethod.Radau: odespy.Radau5,
    ODEMethod.AdamsBashMoulton2: odespy.AdamsBashMoulton2,
    ODEMethod.AdamsBashforth2: odespy.AdamsBashforth2
}


class ODESpySolver(ODESolver):
    """
    ODE solver method implemented on odespy package.
    """

    def __init__(self, func, method):
        self.func = func
        if method in methods.keys():
            self.method = method
        else:
            raise ValueError(f'Method {method} is unavailable.')

        self.initial_condition = None

    def set_initial_condition(self, initial_condition):
        self.initial_condition = initial_condition

    def solve(self, y0, t_span, **kwargs):
        """
        Returns the solver method from odespy package.

        :param func: function with ODE system.
        :return: an instance of odeSolver

        """
        if self.method == ODEMethod.AdamsBashforth2:
            solver = methods[self.method](self.func, method='bdf')
        else:
            solver = methods[self.method](self.func)

            # update default parameters
        solver.nsteps = SolverConfigurations.N_STEPS
        solver.atol = SolverConfigurations.ABSOLUTE_TOL
        solver.rtol = SolverConfigurations.RELATIVE_TOL
        solver.set_initial_condition(y0)
        sol, _ = solver.solve()
        return sol



from collections import namedtuple
from typing import Type, Union, TYPE_CHECKING

# TODO: this module largely depends on pandas dataframes. Should it be set as package requirement?
# noinspection PyPackageRequirements
from pandas import Series, DataFrame, concat

from mewpy.util.constants import ModelConstants
from .solution import ModelSolutionInterface

if TYPE_CHECKING:
    from mewpy.model import Model, MetabolicModel, RegulatoryModel


# TODO: methods stubs and type hinting
class PolymorphicSolution:

    @classmethod
    def from_solver(cls: Type['ModelSolution'],
                    method,
                    solution,
                    **kwargs):
        return cls(method=method,
                   x=solution.values,
                   objective_value=solution.fobj,
                   status=solution.status.value.lower(),
                   reduced_costs=solution.reduced_costs,
                   shadow_prices=solution.shadow_prices,
                   **kwargs)


# TODO: methods stubs and type hinting
class ModelSolution(ModelSolutionInterface, PolymorphicSolution):

    def __init__(self,
                 method,
                 x,
                 objective_value,
                 status,
                 objective_direction='maximize',
                 reduced_costs=None,
                 shadow_prices=None,
                 model=None,
                 simulator=None,
                 tol=ModelConstants.TOLERANCE):

        if not method:
            method = 'fba'

        if not x:
            x = {}

        if not objective_value:
            objective_value = 0.0

        if not status:
            status = 'feasible'

        if not objective_direction:
            objective_direction = 'maximize'

        if not reduced_costs:
            reduced_costs = {}

        if not shadow_prices:
            shadow_prices = {}

        self._method = method
        self._x = x
        self._objective_value = objective_value
        self._status = status
        self._objective_direction = objective_direction
        self._reduced_costs = reduced_costs
        self._shadow_prices = shadow_prices
        self._model = model
        self._simulator = simulator
        self.tol = tol

    @staticmethod
    def _filter_mid_term_variables(x, model):

        if model is None:
            return x

        new_x = {}

        for variable, value in x.items():

            model_variable = model.get(variable, None)

            if model_variable is not None:
                new_x[variable] = value

        return new_x

    @property
    def objective_direction(self):
        return self._objective_direction

    @property
    def method(self):
        return self._method

    @property
    def model(self) -> Union['Model', 'MetabolicModel', 'RegulatoryModel']:
        return self._model

    @property
    def objective(self):

        if self.model:

            if hasattr(self.model, 'objective'):
                return ' + '.join([obj.id for obj in self.model.objective])

        return None

    @property
    def objective_value(self):
        return self._objective_value

    @property
    def simulator(self):
        return self._simulator

    @property
    def status(self):
        return self._status

    @property
    def x(self):
        return self._filter_mid_term_variables(x=self._x, model=self.model)

    @property
    def shadow_prices(self):
        return self._filter_mid_term_variables(x=self._shadow_prices, model=self.model)

    @property
    def reduced_costs(self):
        return self._filter_mid_term_variables(x=self._reduced_costs, model=self.model)

    def _get_variable_info(self, variable):

        identifier = variable.id

        x = self._x.get(variable.id, None)

        v_type = ', '.join(variable.types)

        return identifier, v_type, x

    def _metabolic_frame(self):

        results = {}

        for variable in self.model.yield_reactions():
            _id, v_type, x = self._get_variable_info(variable)

            results[_id] = (variable.id, v_type, x)

        return DataFrame.from_dict(results,
                                   orient='index',
                                   columns=['reaction', 'variable type', 'flux'])

    def _regulatory_frame(self):

        results = {}

        for variable in self.model.yield_regulators():
            _id, v_type, x = self._get_variable_info(variable)

            results[_id] = (variable.id, v_type, x)

        for variable in self.model.yield_targets():
            _id, v_type, x = self._get_variable_info(variable)

            results[_id] = (variable.id, v_type, x)

        return DataFrame.from_dict(results,
                                   orient='index',
                                   columns=['regulatory variable',
                                            'variable type',
                                            'expression coefficient'])

    def _metabolic_environmental_conditions_frame(self):

        results = {}

        for variable in self.model.yield_exchanges():

            if variable.is_reaction():
                _id, v_type, x = self._get_variable_info(variable)

                metabolite = next(iter(variable.metabolites.keys()))

                lb, ub = variable.bounds

                results[_id] = (_id, v_type, metabolite, lb, ub, x)

        return DataFrame.from_dict(results,
                                   orient='index',
                                   columns=['exchange', 'variable type', 'metabolite',
                                            'lower bound', 'upper bound',
                                            'flux'])

    def _regulatory_environmental_conditions_frame(self):

        results = {}

        for variable in self.model.yield_environmental_stimuli():

            if variable.is_regulator():
                _id, v_type, x = self._get_variable_info(variable)

                lb, ub = variable.coefficient.bounds

                results[_id] = (_id, v_type, lb, ub, x)

        return DataFrame.from_dict(results,
                                   orient='index',
                                   columns=['regulatory variable', 'variable type',
                                            'minimum coefficient', 'maximum coefficient',
                                            'expression coefficient'])

    def _regulatory_inputs_frame(self):

        results = {}

        for variable in self.model.yield_environmental_stimuli():

            _id, v_type, x = self._get_variable_info(variable)

            if variable.is_regulator():
                results[_id] = (_id, v_type, x)

        return DataFrame.from_dict(results,
                                   orient='index',
                                   columns=['regulator', 'variable type', 'expression coefficient'])

    def _regulatory_outputs_frame(self):

        results = {}

        for variable in self.model.yield_targets():
            _id, v_type, x = self._get_variable_info(variable)

            results[_id] = (_id, v_type, x)

        return DataFrame.from_dict(results,
                                   orient='index',
                                   columns=['target', 'variable type', 'expression coefficient'])

    def _metabolic_inputs_frame(self):

        results = {}

        for variable in self.model.yield_exchanges():

            _id, v_type, x = self._get_variable_info(variable)

            if variable.is_reaction() and x < -self.tol:
                metabolite = next(iter(variable.metabolites.keys()))

                results[_id] = (_id, v_type, metabolite, x)

        return DataFrame.from_dict(results,
                                   orient='index',
                                   columns=['reaction', 'variable type', 'metabolite', 'flux'])

    def _metabolic_outputs_frame(self):

        results = {}

        for variable in self.model.yield_exchanges():

            _id, v_type, x = self._get_variable_info(variable)

            if variable.is_reaction() and x > self.tol:
                metabolite = next(iter(variable.metabolites.keys()))

                results[_id] = (_id, v_type, metabolite, x)

        return DataFrame.from_dict(results,
                                   orient='index',
                                   columns=['reaction', 'variable type', 'metabolite', 'flux'])

    def _metabolic_summary_frame(self):

        results = {}

        for variable in self.model.yield_exchanges():

            _id, v_type, x = self._get_variable_info(variable)

            if variable.is_reaction() and x < -self.tol:
                metabolite = next(iter(variable.metabolites.keys()))

                results[_id] = (_id, v_type, metabolite, 'input', x)

            if variable.is_reaction() and x > self.tol:
                metabolite = next(iter(variable.metabolites.keys()))

                results[_id] = (_id, v_type, metabolite, 'output', x)

        return DataFrame.from_dict(results,
                                   orient='index',
                                   columns=['reaction', 'variable type', 'metabolite', 'role', 'flux'])

    def _regulatory_summary_frame(self):

        results = {}

        for variable in self.model.yield_environmental_stimuli():

            _id, v_type, x = self._get_variable_info(variable)

            if variable.is_regulator():
                results[_id] = (_id, v_type, 'input', x)

        for variable in self.model.yield_targets():
            _id, v_type, x = self._get_variable_info(variable)

            results[_id] = (_id, v_type, 'output', x)

        return DataFrame.from_dict(results,
                                   orient='index',
                                   columns=['regulatory variable', 'variable type', 'role', 'expression coefficient'])

    def _objective_frame(self):

        return DataFrame([[self.objective_value, self.objective_direction]],
                         index=[self.objective],
                         columns=['value', 'direction'])

    def _get_frame(self, to, dimension):

        dim_to = f'{dimension}_{to}'

        if dim_to == 'regulatory_environmental_conditions':

            return self._regulatory_environmental_conditions_frame()

        elif dim_to == 'metabolic_environmental_conditions':

            return self._metabolic_environmental_conditions_frame()

        elif dim_to == 'regulatory_frame':

            return self._regulatory_frame()

        elif dim_to == 'metabolic_frame':

            return self._metabolic_frame()

        elif dim_to == 'metabolic_inputs':

            return self._metabolic_inputs_frame()

        elif dim_to == 'metabolic_outputs':

            return self._metabolic_outputs_frame()

        elif dim_to == 'regulatory_inputs':

            return self._regulatory_inputs_frame()

        elif dim_to == 'regulatory_outputs':

            return self._regulatory_outputs_frame()

        elif dim_to == 'objective_objective':

            return self._objective_frame()

        elif dim_to == 'regulatory_summary':

            return self._regulatory_summary_frame()

        elif dim_to == 'metabolic_summary':

            return self._metabolic_summary_frame()

        else:
            return DataFrame()

    def _frame_builder(self, to, dimensions):

        fields = tuple(dimensions) + ('frame',)

        Frame = namedtuple('Frame', fields)

        frames = [self._get_frame(to=to, dimension=dimension)
                  for dimension in dimensions]

        if frames:
            frame = concat(frames,
                           axis=1,
                           join='outer',
                           keys=dimensions)

        else:
            frame = DataFrame()

        frames.append(frame)

        # noinspection PyArgumentList
        return Frame(*frames)

    def to_frame(self, dimensions=None, environmental_conditions=False):

        if not dimensions:
            dimensions = self.model.types

        if environmental_conditions:
            return self._frame_builder(to='environmental_conditions', dimensions=dimensions)

        return self._frame_builder(to='frame', dimensions=dimensions)

    def to_series(self):

        return Series(self.x)

    def _summary_builder(self, dimensions):

        fields = ('inputs', 'outputs', 'objective', 'frame') + tuple(dimensions)

        Summary = namedtuple('Summary', fields)

        frames = []

        inputs_frames = [self._get_frame(to='inputs', dimension=dimension)
                         for dimension in dimensions]

        if inputs_frames:
            inputs_frame = concat(inputs_frames,
                                  axis=1,
                                  join='outer',
                                  keys=dimensions)

        else:
            inputs_frame = DataFrame()

        frames.append(inputs_frame)

        outputs_frames = [self._get_frame(to='outputs', dimension=dimension)
                          for dimension in dimensions]

        if outputs_frames:
            outputs_frame = concat(outputs_frames,
                                   axis=1,
                                   join='outer',
                                   keys=dimensions)

        else:
            outputs_frame = DataFrame()

        frames.append(outputs_frame)

        objective_frame = self._get_frame(to='objective', dimension='objective')

        frames.append(objective_frame)

        summary_frames = [self._get_frame(to='summary', dimension=dimension)
                          for dimension in dimensions]

        if summary_frames:
            frame = concat(summary_frames,
                           axis=1,
                           join='outer',
                           keys=dimensions)

        else:
            frame = DataFrame()

        frames.append(frame)
        frames.extend(summary_frames)

        # noinspection PyArgumentList
        return Summary(*frames)

    def to_summary(self, dimensions=None):

        if not dimensions:
            dimensions = self.model.types

        return self._summary_builder(dimensions=dimensions)

    def __repr__(self):
        return f'{self.method} Solution'

    def __str__(self):
        return f'{self.method} {self.status} solution: {self.objective_value}'



from typing import Any, Union, Type, TYPE_CHECKING, List, Set, Dict

from mewpy.util.history import HistoryManager, recorder
from mewpy.util.serialization import serialize, Serializer
from mewpy.mew.lp import Notification

# Preventing circular dependencies that only happen due to type checking
if TYPE_CHECKING:
    from mewpy.model import MetabolicModel, RegulatoryModel
    from mewpy.mew.variables import Gene, Interaction, Metabolite, Reaction, Regulator, Target
    from mewpy.mew.lp import LinearProblem


class MetaModel(type):
    factories = {}

    def __new__(mcs, name, bases, attrs, **kwargs):

        # if it is the model factory, only registration is done
        factory = kwargs.get('factory', False)

        if factory:
            cls = super(MetaModel, mcs).__new__(mcs, name, bases, attrs)

            MetaModel.factories[name] = cls

            return cls

        # Dynamic typing being used. In this case, a proper name and model type must be provided
        dynamic = kwargs.get('dynamic', False)
        if dynamic:
            names = [base.model_type for base in bases]

            name = ''.join([name.title() for name in names])
            name += 'Model'

            kwargs['model_type'] = '-'.join(names)

        # The model type is always added to the subclasses. If it is not given upon subclass creation,
        # the subclass name is to be used
        model_type = kwargs.get('model_type', name.lower())
        attrs['model_type'] = model_type

        return super(MetaModel, mcs).__new__(mcs, name, bases, attrs)

    def __init__(cls, name, bases, attrs, **kwargs):

        super().__init__(name, bases, attrs)

        factory = kwargs.get('factory', False)
        if factory:
            # collection of all containers that must be serialized for a particular class or subclass
            containers = cls.get_serializable_containers(attrs)

            attrs['_containers_registry']['model'] = containers

            # Skip further building of the Model factory
            return

        # Dynamic typing being used. In this case, all children have already been constructed, so everything can be
        # skipped
        dynamic = kwargs.get('dynamic', False)
        if dynamic:
            return

        # Several attributes and methods must be added automatically to the Model factory children based on the
        # model type.
        model_type = attrs['model_type']

        for base in bases:

            factory = MetaModel.factories.get(base.__name__)

            if factory and hasattr(cls, 'model_type'):

                # if some class inherits from the Model factory, it must be registered in the factory for type
                # checking
                if cls.model_type not in factory.get_registry():
                    raise TypeError(f'{cls.model_type} does not inherit from Model')

                # If set otherwise upon class creation, all subclasses of the Model factory will contribute to
                # their parent factory with an alternative initializer. The polymorphic constructor, e.g. from_{
                # model_type}, is added to the Model factory
                constructor = kwargs.get('constructor', True)

                if constructor:
                    # noinspection PyProtectedMember
                    model_type_initializer = Model._from(cls)
                    setattr(factory, f'from_{model_type}', model_type_initializer)

                # If set otherwise upon class creation, all subclasses of the Model factory will contribute to
                # their parent factory with a declared type checker. The type checker, e.g. is_{model_type},
                # is added to the Model factory
                checker = kwargs.get('checker', True)

                if checker:
                    # noinspection PyProtectedMember
                    model_type_checker = Model._is(model_type)
                    setattr(factory, f'is_{model_type}', model_type_checker)

                # collection of all containers that must be serialized for a particular class or subclass
                containers = cls.get_serializable_containers(attrs)

                Model.register_containers(containers, cls)

    @staticmethod
    def get_serializable_containers(attrs):

        # property getter is marked with serialize and deserialize attribute by the serialize decorator. If
        # deserialization is not to be used, deserialize should be marked with None

        containers = {}

        for name, method in attrs.items():

            if hasattr(method, 'fget'):

                if hasattr(method.fget, 'serialize') and hasattr(method.fget, 'deserialize') and hasattr(method.fget,
                                                                                                         'pickle'):
                    containers[name] = (method.fget.serialize, method.fget.deserialize, method.fget.pickle)

        return containers


# TODO: methods stubs and type hinting
class Model(Serializer, metaclass=MetaModel, factory=True):
    # -----------------------------------------------------------------------------
    # Factory management
    # -----------------------------------------------------------------------------

    _registry = {}
    _containers_registry = {'model': {}}

    def __init_subclass__(cls, **kwargs):

        super(Model, cls).__init_subclass__(**kwargs)

        # the child type
        model_type = getattr(cls, 'model_type', cls.__name__.lower())

        cls.register_type(model_type, cls)

    @staticmethod
    def get_registry():
        return Model._registry.copy()

    @staticmethod
    def get_containers_registry():
        return Model._containers_registry.copy()

    @staticmethod
    def register_type(model_type, child):

        Model._registry[model_type] = child

    # -----------------------------------------------------------------------------
    # Serialization
    # -----------------------------------------------------------------------------

    @staticmethod
    def register_containers(containers, child):

        if hasattr(child, 'model_type'):

            Model._containers_registry[child.model_type] = containers

        elif child is Model:
            Model._containers_registry['model'] = containers

    @property
    def containers(self):

        class_containers = self.get_containers_registry()

        containers = class_containers['model'].copy()

        for model_type in self.types:

            for name, (serialize_name, deserialize_name, pickle_name) in class_containers[model_type].items():
                containers[name] = (serialize_name, deserialize_name, pickle_name)

        return containers

    # -----------------------------------------------------------------------------
    # Factory polymorphic constructor
    # -----------------------------------------------------------------------------

    @classmethod
    def factory(cls, *args: str) -> Union[Type['Model'],
                                          Type['MetabolicModel'],
                                          Type['RegulatoryModel']]:

        if not args:
            args = ()

        registry = cls.get_registry()

        types = tuple([registry[name] for name in args])

        if len(types) == 1:
            return types[0]

        _Model = MetaModel('Model', types, {}, dynamic=True)

        return _Model

    # -----------------------------------------------------------------------------
    # Factory polymorphic initializer
    # -----------------------------------------------------------------------------

    @classmethod
    def from_types(cls, types: List[str], **kwargs) -> Union['Model',
                                                             'MetabolicModel',
                                                             'RegulatoryModel']:
        ModelType = cls.factory(*types)
        return ModelType(**kwargs)

    # -----------------------------------------------------------------------------
    # Factory helper methods
    # -----------------------------------------------------------------------------

    @staticmethod
    def _from(cls):

        def from_(identifier, **kwargs):
            return cls(identifier, **kwargs)

        return from_

    @staticmethod
    def _is(model_type):

        def is_(self):
            return self.is_a(model_type)

        return is_

    # -----------------------------------------------------------------------------
    # Base initializer
    # -----------------------------------------------------------------------------

    def __init__(self,
                 identifier: Any,
                 name: str = None):

        """

        The model is the most base type of a given model type, such as metabolic model or regulatory model.
        See also model.MetabolicModel and model.RegulatoryModel for concrete implementations of a model type.

        The model object is the factory for several model types.
        This model object provides all batteries to create new model types and to manage the model types' inheritance.

        The factory type assists with systematic tasks such as attributes and containers serialization,
        history management, simulator observation, among others.

        For that, attributes, containers and inheritance are registered in the factory. The MetaModel manages
        both the factory and the model types derived from the factory.

        :param identifier: identifier,  e.g. iMC1010
        :param name: the name of the model
        """

        self._check_inheritance()

        if not identifier:
            identifier = ''

        if not name:
            name = identifier

        self._id = ''
        self._name = ''
        self._simulators = []
        self._types = set()

        # History: reversible changes to the model
        self._contexts = []
        self._history = HistoryManager()

        self._id = identifier
        self.name = name

        # models share different compartments. Although it is not the most subtle solution, name mangling is used to
        # store the specific compartments of each child. An alternative would be to check for shared attributes
        # between all children and perform this solution for them
        self.__compartments = {}

    def _check_inheritance(self):

        registry = self.get_registry()

        for model_type in self.types:
            if model_type not in registry:
                raise ValueError(f'{model_type} is not registered as subclass of {self.__class__.__name__}')

    # -----------------------------------------------------------------------------
    # Built-in
    # -----------------------------------------------------------------------------

    def __str__(self):
        return self.name

    def __repr__(self):
        if self.types:
            types = [v_type.title() for v_type in self.types]

            return ', '.join(types) + f': {self.id}'

        return f'Model: {self.id}'

    # -----------------------------------------------------------------------------
    # Variable type manager
    # -----------------------------------------------------------------------------

    @serialize('types', None, None)
    @property
    def types(self) -> Set[str]:
        return set()

    # -----------------------------------------------------------------------------
    # Static attributes
    # -----------------------------------------------------------------------------

    @serialize('id', None, '_id')
    @property
    def id(self) -> Any:
        return self._id

    @serialize('name', 'name', '_name')
    @property
    def name(self) -> str:
        return self._name

    @property
    def simulators(self) -> List['LinearProblem']:
        return self._simulators

    # -----------------------------------------------------------------------------
    # Static attributes setters
    # -----------------------------------------------------------------------------

    @name.setter
    @recorder
    def name(self, value: str):

        if not value:
            value = ''

        self._name = value

    # -----------------------------------------------------------------------------
    # Operations/Manipulations
    # -----------------------------------------------------------------------------

    def get(self, identifier: Any, default=None) -> Union['Gene',
                                                          'Interaction',
                                                          'Metabolite',
                                                          'Reaction',
                                                          'Regulator',
                                                          'Target']:
        return default

    # add and remove are just registered here to avoid type checking errors

    def add(self, variables, *types, comprehensive=True, history=True):

        pass

    def remove(self, variables, *types, remove_orphans=False, history=True):

        pass

    def update(self, name: str = None):

        if name:
            self.name = name

    def _add_variable_to_container(self, variable, container):

        if isinstance(container, str):
            container = getattr(self, container, container)

        if variable.id not in container:
            container[variable.id] = variable

            if variable.model is not self:
                variable.model = self
                # noinspection PyProtectedMember
                variable._model_ref = 1

            elif variable.model is self:
                # noinspection PyProtectedMember
                variable._model_ref += 1

    def _remove_variable_from_container(self, variable, container):

        if isinstance(container, str):
            container = getattr(self, container, container)

        if variable.id in container:

            del container[variable.id]

            # noinspection PyProtectedMember
            if variable._model_ref < 2:
                variable.model = None

            elif variable._model_ref > 1:
                # noinspection PyProtectedMember
                variable._model_ref -= 1

    # -----------------------------------------------------------------------------
    # Simulators observer pattern
    # -----------------------------------------------------------------------------

    def attach(self, simulator: 'LinearProblem'):

        self.simulators.append(simulator)

    def detach(self, simulator):

        self.simulators.remove(simulator)

    def notify(self, notification: Notification):

        for simulator in self.simulators:
            simulator.notification(notification)

    # -----------------------------------------------------------------------------
    # History manager
    # -----------------------------------------------------------------------------

    # The history manager follows the command pattern
    @property
    def contexts(self) -> List[HistoryManager]:
        return self._contexts

    @property
    def history(self) -> HistoryManager:

        if self.contexts:
            return self.contexts[-1]

        return self._history

    def undo(self):

        self.history.undo()

    def redo(self):

        self.history.redo()

    def clean_history(self):

        self._history = HistoryManager()

    def clean_context(self, index=None):

        if not index:
            index = -1

        self.contexts[index] = HistoryManager()

    def reset(self):

        self.history.reset()

    def restore(self):

        self.history.restore()

    def __enter__(self):

        self.contexts.append(HistoryManager())

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):

        context = self.contexts.pop()
        context.reset()

    # -----------------------------------------------------------------------------
    # Abstract implementations
    # -----------------------------------------------------------------------------

    # The following polymorphic initializers are just registered here to avoid type checking errors
    @classmethod
    def from_metabolic(cls,
                       identifier: Any,
                       name: str = None,
                       compartments: Dict[str, str] = None,
                       genes: Dict[str, 'Gene'] = None,
                       metabolites: Dict[str, 'Metabolite'] = None,
                       objective: Dict['Reaction', Union[float, int]] = None,
                       reactions: Dict[str, 'Reaction'] = None) -> 'MetabolicModel':
        ...

    @classmethod
    def from_regulatory(cls,
                        identifier: Any,
                        name: str = None,
                        compartments: Dict[str, str] = None,
                        interactions: Dict[str, 'Interaction'] = None,
                        regulators: Dict[str, 'Regulator'] = None,
                        targets: Dict[str, 'Target'] = None) -> 'RegulatoryModel':
        ...

    # The following type checkers are just registered here to avoid type checking errors
    def is_metabolic(self) -> bool:
        ...

    def is_regulatory(self) -> bool:
        ...

    # -----------------------------------------------------------------------------
    # Type checker
    # -----------------------------------------------------------------------------

    def is_a(self, model_type) -> bool:
        if model_type in self.types:
            return True

        return False

    # -----------------------------------------------------------------------------
    # Common attributes - Name mangling
    # -----------------------------------------------------------------------------

    # TODO: perhaps name mangling should be managed by the metaclass, as more common attributes can occur
    @property
    def compartments(self):

        return self.__compartments

    # -----------------------------------------------------------------------------
    # Helper methods
    # -----------------------------------------------------------------------------

    @staticmethod
    def _get_orphans(to_remove, first_container, second_container):

        orphans = set()

        for variable in to_remove:

            container_iter = getattr(variable,
                                     f'yield_{first_container}',
                                     lambda: [getattr(variable, f'{first_container}')])

            for variable_2 in container_iter():

                remove_variable = True

                container_iter2 = getattr(variable_2,
                                          f'yield_{second_container}',
                                          lambda: [getattr(variable_2, f'{second_container}')])

                for variable_3 in container_iter2():

                    if variable_3 not in to_remove:
                        remove_variable = False

                if remove_variable:
                    orphans.add(variable_2)

        return orphans


def build_model(types, kwargs):
    return Model.from_types(types, **kwargs)



from typing import Union, TYPE_CHECKING

if TYPE_CHECKING:

    from io import TextIOWrapper

    from mewpy.io.engines.engine import Engine

    try:
        # noinspection PyPackageRequirements
        from cobra import Model as Cobra_Model

    except ImportError:
        Cobra_Model = str

    try:
        # noinspection PyPackageRequirements
        from reframed import ReframedModel as Reframed_Model

    except ImportError:
        Reframed_Model = str


class Builder:

    def __init__(self):
        """
        Builder base for an engine wrapper of any kind.
        That is, the builder does not provide anything by itself.
        Instead, this class is to be used by the reader and writer, which are just wrappers of engines.

        # Wrapping is accomplished using the composition pattern

        Engines, on the other hand, do the hard work of opening, parsing and reading/writing a specific file or model.
        Basically, engines are, actually, the builders of models or files

        """
        self._engine: 'Engine' = None

        # useful for the context manager and the read or write methods
        self._context = False

    @property
    def engine(self) -> 'Engine':
        return self._engine

    @property
    def io(self) -> Union[str, 'TextIOWrapper', 'Cobra_Model', 'Reframed_Model']:
        return self.engine.io

    @property
    def config(self) -> dict:
        return self.engine.config

    @property
    def warnings(self) -> tuple:
        return tuple(self.engine.warnings)



from math import inf

from ...solvers import solver_instance
from ...simulation import get_simulator
from ...simulation.simulation import Simulator
from .. import Preprocessing, ExpressionSet


def GIMME(model, expr, biomass=None, condition=0, cutoff=0.25, growth_frac=0.9,
          constraints=None, parsimonious=False, **kwargs):
    """ Run a GIMME simulation [1]_.

    Arguments:
        model: a REFRAMED or COBRApy model or a MEWpy Simulator.
        expr (ExpressionSet): transcriptomics data.
        biomass: the biomass reaction identifier
        condition: the condition to use in the simulation\
            (default:0, the first condition is used if more than one.)
        cutoff (int): percentile cuttof (default: 0.25).
        growth_frac (float): minimum growth requirement (default: 0.9)
        constraints (dict): additional constraints
        parsimonious (bool): compute a parsimonious solution (default: False)

    Returns:
        Solution: solution

    References
    ----------
    .. [1] Becker, S. and Palsson, B. O. (2008).
           Context-specific metabolic networks are consistent with experiments.
           PLoS Computational Biology, 4(5), e1000082.
           doi:10.1371/journal.pcbi.1000082
    """
    if isinstance(model, Simulator):
        sim = model
    else:
        sim = get_simulator(model)

    if isinstance(expr, ExpressionSet):
        pp = Preprocessing(sim, expr)
        coeffs, _ = pp.percentile(condition, cutoff=cutoff)
        solver = solver_instance(sim)
    else:
        coeffs = expr
        solver = solver_instance(sim)

    # TODO: improve this on the simulator side
    if biomass is None:
        try:
            biomass = list(sim.objective.keys())[0]
        except Exception:
            raise ValueError(
                "A biomass reaction identifier is required or "
                "needs to be set as the model objective")

    wt_solution = sim.simulate(constraints=constraints)

    if not constraints:
        constraints = {}
    # add growth constraint
    constraints[biomass] = (growth_frac * wt_solution.fluxes[biomass], inf)

    # make model irreversible
    for r_id in sim.reactions:
        lb, _ = sim.get_reaction_bounds(r_id)
        if lb < 0:
            pos, neg = r_id + '+', r_id + '-'
            solver.add_variable(pos, 0, inf, update=False)
            solver.add_variable(neg, 0, inf, update=False)
    solver.update()

    for r_id in sim.reactions:
        lb, _ = sim.get_reaction_bounds(r_id)
        if lb < 0:
            pos, neg = r_id + '+', r_id + '-'
            solver.add_constraint(
                'c' + pos, {r_id: -1, pos: 1}, '>', 0, update=False)
            solver.add_constraint(
                'c' + neg, {r_id: 1, neg: 1}, '>', 0, update=False)
    solver.update()

    objective = dict()
    for r_id, val in coeffs.items():
        lb, _ = sim.get_reaction_bounds(r_id)
        if lb < 0:
            pos, neg = r_id + '+', r_id + '-'
            objective[pos] = val
            objective[neg] = val
        else:
            objective[r_id] = val

    solution = solver.solve(objective, minimize=True, constraints=constraints)

    if parsimonious:
        pre_solution = solution

        solver.add_constraint('obj', objective, '=', pre_solution.fobj)
        objective = dict()

        for r_id in sim.reactions:
            lb, _ = sim.get_reaction_bounds(r_id)
            if lb < 0:
                pos, neg = r_id + '+', r_id + '-'
                objective[pos] = 1
                objective[neg] = 1
            else:
                objective[r_id] = 1

        solution = solver.solve(objective, minimize=True,
                                constraints=constraints)
        solver.remove_constraint('obj')
        solution.pre_solution = pre_solution

    for r_id in sim.reactions:
        lb, _ = sim.get_reaction_bounds(r_id)
        if lb < 0:
            pos, neg = r_id + '+', r_id + '-'
            del solution.values[pos]
            del solution.values[neg]

    return solution



import os
from functools import partial
from math import ceil
from typing import Union, TYPE_CHECKING

from mewpy.mew.algebra import Expression, Symbol, Or, And, NoneAtom, Float
from mewpy.model import RegulatoryModel, MetabolicModel
from mewpy.io.dto import DataTransferObject, VariableRecord, History, FunctionTerm, CompartmentRecord
from mewpy.util.constants import ModelConstants

from .engine import Engine
from .engines_utils import (build_symbolic,
                            ASTNODE_BOOLEAN_VALUES, ASTNODE_RELATIONAL_OPERATORS,
                            ASTNODE_NAME, ASTNODE_BOOLEAN_OPERATORS, ASTNODE_VALUES,
                            pattern_notes,
                            f_id, fs_id,
                            F_GENE, F_SPECIE, F_REACTION, F_SPECIE_REV, F_GENE_REV, F_REACTION_REV, F_TRANSITION,
                            F_TRANSITION_REV,
                            convert_fbc, get_sbml_doc_to_write, get_sbml_doc_to_read,
                            UNIT_ID, UNITS,
                            add_sbml_parameter, LOWER_BOUND_ID, UPPER_BOUND_ID, ZERO_BOUND_ID,
                            BOUND_MINUS_INF, BOUND_PLUS_INF,
                            SBO_DEFAULT_FLUX_BOUND, SBO_FLUX_BOUND,
                            get_sbml_lb_id, get_sbml_ub_id, write_sbml_doc,
                            set_math, set_gpr,
                            expression_warning, sbml_warning)

if TYPE_CHECKING:
    from mewpy.model import Model, MetabolicModel, RegulatoryModel


class RegulatorySBML(Engine):

    def __init__(self, io, config, model=None):

        super().__init__(io, config, model)

    @property
    def model_type(self):
        return 'regulatory'

    @property
    def model(self):

        if self._model is None:

            identifier = self.dto.id

            if not identifier:
                identifier = self.get_identifier()

            return RegulatoryModel(identifier=identifier)

        return self._model

    @staticmethod
    def _parse_coefficients_note(notes):

        if not notes:
            return set()

        coefficients = set()

        for row in notes.split('\n'):

            row = row.strip()

            if row.startswith('<p>') and row.endswith('</p>'):
                st, level = row.split(':')

                st = st.strip('<p>')
                level = level.strip('</p>')

                if 'state' in st.lower():
                    st = int(st.lower().replace('state', '').strip())

                lb = ModelConstants.REACTION_LOWER_BOUND
                ub = ModelConstants.REACTION_UPPER_BOUND

                level = level.replace('+inf', f'{ub}').replace('-inf', f'{lb}')

                min_coef, max_coef = level.split(',')

                min_coef = float(min_coef[1:])
                coefficients.add(min_coef)

                max_coef = float(max_coef[0:-1])
                coefficients.add(max_coef)

        return coefficients

    @staticmethod
    def _parse_initial_level_note(notes):

        if not notes:
            return

        for row in notes.split('\n'):

            row = row.strip()

            if row.startswith('<p>') and row.endswith('</p>'):
                parameter, level = row.split(':')

                parameter = parameter.strip('<p>')
                level = level.strip('</p>')

                if 'initial' in parameter.lower():
                    return float(level.replace(' ', ''))

    def _update_qual_species_coefficients(self, qual_species, value):

        # all species are created with the following coefficients: {0:0, 1: maximum level}.
        # but intermediate states can be encoded into the transitions and thus added to the coefficients dict

        value = float(value)

        if qual_species not in self.dto.variables:
            raise SyntaxError

        coefficients = self.dto.variables[qual_species].coefficients

        if min(coefficients) > value > max(coefficients):
            raise SyntaxError

        coefficients.add(value)

    def _parse_values(self, ast_node, ast_node_qual_species, inputs_thresholds):

        node_type = ast_node.getType()

        if node_type in ASTNODE_BOOLEAN_VALUES:

            return ASTNODE_BOOLEAN_VALUES[node_type]

        elif ast_node.getType() == ASTNODE_NAME:

            # In this case, the right child is encoded in the list of inputs

            input_id = fs_id(ast_node.getName(), (F_SPECIE, F_GENE, F_REACTION, F_TRANSITION))
            value = inputs_thresholds.get(input_id, None)

            if value is None:
                raise SyntaxError

            regulator_id = fs_id(ast_node_qual_species.getName(), (F_SPECIE, F_GENE, F_REACTION))
            self._update_qual_species_coefficients(regulator_id, value)

            return Float(value)

        elif ast_node.getType() in ASTNODE_VALUES:

            # In this case, the right child is encoded in the regulators current coefficients

            value = ast_node.getValue()

            if value is None:
                raise SyntaxError

            regulator_id = fs_id(ast_node_qual_species.getName(), (F_SPECIE, F_GENE, F_REACTION))
            self._update_qual_species_coefficients(regulator_id, value)

            return Float(value)

        else:
            raise SyntaxError

    def _parse_symbols(self, ast_node):

        ast_node_id = fs_id(ast_node.getName(), (F_SPECIE, F_GENE, F_REACTION))

        if ast_node_id not in self.dto.variables:
            raise SyntaxError

        return Symbol(value=ast_node_id)

    def _parse_math_node(self, ast_node, inputs_thresholds):

        node_type = ast_node.getType()

        if node_type in ASTNODE_BOOLEAN_OPERATORS:

            Operator = ASTNODE_BOOLEAN_OPERATORS[node_type]
            variables = []

            for child in range(ast_node.getNumChildren()):
                variable = self._parse_math_node(ast_node.getChild(child), inputs_thresholds)

                variables.append(variable)

            return Operator(variables=variables)

        elif node_type in ASTNODE_RELATIONAL_OPERATORS:

            Operator = ASTNODE_RELATIONAL_OPERATORS[node_type]
            symbol = self._parse_symbols(ast_node.getChild(0))
            value = self._parse_values(ast_node.getChild(1), ast_node.getChild(0), inputs_thresholds)

            return Operator(variables=[symbol, value])

        elif node_type == ASTNODE_NAME:

            return self._parse_symbols(ast_node)

        else:

            raise SyntaxError

    def parse_math_node(self, ast_node, inputs_thresholds):

        """
        Reads and parses a node of type math ASTNode into an algebraic expression.

        :param ast_node: ASTNode, math ASTNode from functionTerm.getMath()
        :param inputs_thresholds:
        :return: Symbolic
        """

        if not inputs_thresholds:
            inputs_thresholds = {}

        try:

            symbolic = self._parse_math_node(ast_node, inputs_thresholds)

        except SyntaxError:

            self.warnings.append(partial(expression_warning,
                                         f'{ast_node} cannot be parsed. Assigning empty expression instead'))

            symbolic = NoneAtom()

        return symbolic

    def get_identifier(self):

        if os.path.exists(self.io):
            _, identifier = os.path.split(self.io)
            return os.path.splitext(identifier)[0]

        return 'model'

    def _open_to_read(self):

        self._dto = DataTransferObject()

        # -----------------------------------------------------------------------------
        # Doc, Model and Qual Plugin
        # -----------------------------------------------------------------------------

        self.dto.doc = get_sbml_doc_to_read(self.io)

        self.dto.model = self.dto.doc.getModel()

        if self.dto.model is None:
            raise OSError(f'{self.io} is not a valid input. Model SBML section is missing. '
                          f'Provide a correct path or file handler')

        self.dto.qual_plugin = self.dto.model.getPlugin('qual')

        if self.dto.qual_plugin is None:
            raise OSError(f'Although {self.io} is a valid SBML file, the qual plugin was not detected. Thus, '
                          f'regulatory interactions cannot be determined')

        identifier = self.dto.model.getIdAttribute()

        if not identifier:
            self.warnings.append(partial(sbml_warning, 'Model identifier is not encoded in the SBML file'))

            identifier = self.get_identifier()

        self.dto.id = identifier
        self.dto.name = self.dto.model.getName()

    def _open_to_write(self):

        self._dto = DataTransferObject()

        # -----------------------------------------------------------------------------
        # Doc, Model and FBC Plugin
        # -----------------------------------------------------------------------------

        self.dto.doc = get_sbml_doc_to_write(self.io,
                                             level=3,
                                             version=1,
                                             packages=('qual',),
                                             packages_version=(1,),
                                             packages_required=(True,),
                                             sbo_term=False)

        if self.dto.model is None:

            self.dto.model = self.dto.doc.createModel()

        else:

            self.dto.model = self.dto.doc.getModel()

        self.dto.qual_plugin = self.dto.model.getPlugin('qual')

        if self.model.id is not None:
            self.dto.model.setId(self.model.id)
            self.dto.model.setMetaId('meta_' + self.model.id)

        else:
            self.dto.model.setMetaId('meta_model')

        if self.model.name is not None:
            self.dto.model.setName(self.model.name)

    def open(self, mode='r'):

        if mode == 'r':

            return self._open_to_read()

        elif mode == 'w':

            return self._open_to_write()

        else:
            raise ValueError(f'{mode} mode is not recognized. Try one of the following: r, w')

    def parse(self):

        if self.dto is None:
            raise OSError('SBML file is not open')

        if self.dto.id is None:
            raise OSError('SBML file is not open')

        if self.dto.doc is None:
            raise OSError('SBML file is not open')

        if self.dto.model is None:
            raise OSError(f'SBML file is not open')

        if self.dto.qual_plugin is None:
            raise OSError(f'SBML file is not open')

        self.dto.level = self.dto.model.getLevel()
        self.dto.version = self.dto.model.getVersion()

        has_history = self.dto.model.isSetModelHistory()

        if has_history:

            history = self.dto.model.getModelHistory()

            created = None
            if history.isSetCreatedDate():
                created = history.getCreatedDate()

            creators = history.getListCreators()

            self.dto.history = History(data=created, creators=creators)

        # -----------------------------------------------------------------------------
        # Compartments
        # -----------------------------------------------------------------------------
        for compartment in self.dto.model.getListOfCompartments():
            self.dto.compartments[compartment.getIdAttribute()] = CompartmentRecord(id=compartment.getIdAttribute(),
                                                                                    name=compartment.getName())

        # -----------------------------------------------------------------------------
        # Qualitative Species
        # -----------------------------------------------------------------------------
        for qual_species in self.dto.qual_plugin.getListOfQualitativeSpecies():

            identifier = fs_id(qual_species.getIdAttribute(), (F_SPECIE, F_GENE, F_REACTION))
            name = qual_species.getName()

            if not name:
                name = identifier

            aliases = {identifier, name}
            compartment = qual_species.getCompartment()

            constant = qual_species.getConstant()

            # coefficients can be encoded into the notes section. It is not advised though
            notes = qual_species.getNotesString()

            coefficients = self._parse_coefficients_note(notes)

            if coefficients:

                self.warnings.append(partial(sbml_warning, f'Are the {identifier} coefficients encoded in the notes '
                                                           f'section? Coefficients must be hard coded during the '
                                                           f'math nodes of each transition.'))

                maximum_level = max(coefficients)
                minimum_level = min(coefficients)

            else:

                # coefficients might be updated later during transition parsing
                maximum_level = float(qual_species.getMaxLevel())
                minimum_level = 0.0

                coefficients = {minimum_level, maximum_level}

            # finding the active coefficient
            if qual_species.isSetInitialLevel():
                active_coefficient = float(qual_species.getInitialLevel())

            else:

                active_coefficient = self._parse_initial_level_note(notes)

                if active_coefficient is None:
                    self.warnings.append(partial(sbml_warning, f'{identifier} initial level was not found. '
                                                               f'Setting active/initial coefficient to the '
                                                               f'minimum value'))

                    active_coefficient = 0.0

            if minimum_level > active_coefficient > maximum_level:

                raise ValueError(f'Initial level is higher/lower than the minimum/maximum level for the {identifier} '
                                 f'qual species')

            else:

                coefficients.add(active_coefficient)

            variable = VariableRecord(id=identifier,
                                      name=name,
                                      aliases=aliases,
                                      compartment=compartment,
                                      constant=constant,
                                      notes=notes,
                                      coefficients=coefficients,
                                      active_coefficient=active_coefficient)

            self.dto.variables[identifier] = variable

        # -----------------------------------------------------------------------------
        # Interactions/List of Transitions
        # -----------------------------------------------------------------------------

        # iterate over all transitions/interactions
        for transition in self.dto.qual_plugin.getListOfTransitions():

            identifier = f_id(transition.getIdAttribute(), F_TRANSITION)
            name = transition.getName()

            if not name:
                name = identifier

            # -----------------------------------------------------------------------------
            # Regulators/List of Inputs
            # -----------------------------------------------------------------------------

            # regulators can be encoded into the list of inputs by the qualitative species or by an input identifier.
            # A regulator associated with an input identifier must also be associated with the qualitative species
            # and sometimes with the threshold level.
            # The threshold level stands for the coefficient of the regulator in a given function term.
            # In this case, the ASTNode math parsing will have to replace the input id the is enclosed by the tag <ci>
            # by the threshold level.

            # A regulator encoded into the list of inputs using only the qualitative species is a simpler approach.
            # In this case, the threshold level is regularly encoded in a <cn> tag of type integer or rational

            # Nevertheless, the second option is more powerful, as we can set multiple non-integer values

            # to be used during functional term parsing
            regulators = {}
            # to be used during functional term parsing
            inputs_thresholds = {}

            for regulator_input in transition.getListOfInputs():

                if not regulator_input.isSetQualitativeSpecies():
                    raise ValueError(f"Qualitative species for {identifier}'s transition input is not set")

                regulator_id = fs_id(regulator_input.getQualitativeSpecies(), (F_SPECIE, F_GENE, F_REACTION))

                regulator_record = self.dto.variables[regulator_id]

                regulators[regulator_id] = regulator_record

                self.dto.regulators[regulator_id] = regulator_record

                self.variables[regulator_id].add('regulator')

                # input identifier
                if regulator_input.isSetId():

                    input_id = fs_id(regulator_input.getIdAttribute(), (F_SPECIE, F_GENE, F_REACTION, F_TRANSITION))

                else:
                    input_id = regulator_id

                # input threshold level
                if regulator_input.isSetThresholdLevel():
                    input_threshold = regulator_input.getThresholdLevel()

                    inputs_thresholds[input_id] = input_threshold

            if inputs_thresholds:
                self.warnings.append(partial(sbml_warning, f'{identifier} threshold levels detected. It is '
                                                           f'recommended to encode input levels directly in the math '
                                                           f'node'))

            # -----------------------------------------------------------------------------
            # Targets/List of Outputs
            # -----------------------------------------------------------------------------

            # In the case of outputs/targets list, only the qualitative species really matters
            targets = {}
            for target_output in transition.getListOfOutputs():

                if not target_output.isSetQualitativeSpecies():
                    raise ValueError(f"Qualitative species for {identifier}'s transition output is not set")

                target_id = fs_id(target_output.getQualitativeSpecies(), (F_SPECIE, F_GENE, F_REACTION))

                target_record = self.dto.variables[target_id]

                targets[target_id] = target_record

                self.dto.targets[target_id] = target_record

                self.variables[target_id].add('target')

            # -----------------------------------------------------------------------------
            # Function Terms
            # -----------------------------------------------------------------------------

            # A function term specify how regulators determine the state of the targets
            function_terms = {}

            # -----------------------------------------------------------------------------
            # Default term
            # -----------------------------------------------------------------------------

            # The default term does not have an expression, as it just refers to the base coefficient value
            # for the target.
            default_term = transition.getDefaultTerm()

            if default_term is None:

                self.warnings.append(partial(sbml_warning, f'Default function term '
                                                           f'not set for transition {identifier}.'
                                                           f'Setting default function term of zero.'))

                result_level = 0.0

            else:

                result_level = float(default_term.getResultLevel())

                if result_level is None:
                    result_level = 0.0

                    self.warnings.append(partial(sbml_warning, 'Default function term result level not found. '
                                                               'Setting to zero'))

            function_terms[result_level] = FunctionTerm(id=f'{identifier}_{result_level}',
                                                        symbolic=NoneAtom(),
                                                        coefficient=result_level)

            # -----------------------------------------------------------------------------
            # Math based function terms
            # -----------------------------------------------------------------------------

            # The other function terms define how the outputs/targets can take multiple states.
            # These terms encode the conditions and logic for a given target being active or inactive
            for function_term in transition.getListOfFunctionTerms():

                # result level of the function term
                result_level = function_term.getResultLevel()

                if result_level is None:
                    result_level = 0

                    self.warnings.append(partial(sbml_warning, f'Function term result level not found '
                                                               f'for term in transition {identifier}. '
                                                               f'Setting to zero'))

                # a transition can only have one function term per result level
                if result_level in function_terms:
                    raise ValueError(f'Function term {identifier} with {result_level} result level has already '
                                     f'been used by other function terms')

                # a transition can only have function terms to which the result levels are less or equal than the
                # maximum levels of all its outputs. _update_qual_species_coefficients will check these issues
                for target in targets:
                    self._update_qual_species_coefficients(target, result_level)

                # math node of the function term
                if function_term.isSetMath():

                    math_node = function_term.getMath()

                    # It will parse the symbolic algebraic expression in the ASTMath node. The numeric atoms will be
                    # replaced by the state coefficient according to the inputs' threshold levels or qualitative
                    # species' coefficients
                    symbolic = self.parse_math_node(math_node, inputs_thresholds)

                else:

                    self.warnings.append(partial(sbml_warning, f'Function term math node was'
                                                               f'not set for transition {identifier} with '
                                                               f'{result_level}.'
                                                               f'Setting function term equal to the result level.'))

                    symbolic = NoneAtom()

                function_terms[result_level] = FunctionTerm(id=f'{identifier}_{result_level}',
                                                            symbolic=symbolic,
                                                            coefficient=result_level)

            # -----------------------------------------------------------------------------
            # Interaction record
            # -----------------------------------------------------------------------------

            # Setting an interaction per target/output

            for target in targets.values():
                interaction_id = f'{target.id}_interaction'

                interaction_record = VariableRecord(id=interaction_id,
                                                    name=identifier,
                                                    aliases={interaction_id, identifier, name, target.id},
                                                    target=target,
                                                    function_terms=function_terms,
                                                    regulators=regulators)

                self.dto.interactions[interaction_id] = interaction_record

                self.variables[interaction_id].add('interaction')

    def read(self,
             model: Union['Model', 'MetabolicModel', 'RegulatoryModel'] = None,
             variables=None):

        if not model:
            model: Union['Model', 'MetabolicModel', 'RegulatoryModel'] = self.model

        if not variables:
            variables = self.variables

        if self.dto.id:
            model._id = self.dto.id

        if self.dto.name:
            model.name = self.dto.name

        model.compartments = {compartment.id: compartment.name
                              for compartment in self.dto.compartments.values()}

        processed_vars = set()

        for interaction_id, interaction_record in self.dto.interactions.items():

            target_record = interaction_record.target

            target, warning = target_record.to_variable(model=model,
                                                        types=variables.get(target_record.id, {'target'}),
                                                        name=target_record.name,
                                                        aliases=target_record.aliases,
                                                        coefficients=target_record.coefficients,
                                                        active_coefficient=target_record.active_coefficient)

            if warning:
                self.warnings.append(partial(sbml_warning, warning))

            processed_vars.add(target_record.id)

            regulators_records = interaction_record.regulators

            regulators = {}

            for regulator_id, regulator_record in regulators_records.items():

                regulator, warn = regulator_record.to_variable(model=model,
                                                               types=variables.get(regulator_id, {'regulator'}),
                                                               name=regulator_record.name,
                                                               aliases=regulator_record.aliases,
                                                               coefficients=regulator_record.coefficients,
                                                               active_coefficient=regulator_record.active_coefficient)

                if warn:
                    self.warnings.append(partial(sbml_warning, warn))

                regulators[regulator_id] = regulator

                processed_vars.add(regulator_id)

            regulatory_events = {}

            for func_term in interaction_record.function_terms.values():
                expression_regulators = {symbol.name: regulators[symbol.name]
                                         for symbol in func_term.symbolic.atoms(symbols_only=True)}

                regulatory_events[func_term.coefficient] = Expression(symbolic=func_term.symbolic,
                                                                      variables=expression_regulators)

            interaction, warn = interaction_record.to_variable(model=model,
                                                               types=variables.get(interaction_id, {'interaction'}),
                                                               name=interaction_record.name,
                                                               aliases=interaction_record.aliases,
                                                               target=target,
                                                               regulatory_events=regulatory_events)

            if warn:
                self.warnings.append(partial(sbml_warning, warn))

            model.add(interaction, 'interaction', comprehensive=True)

        if len(processed_vars) != len(self.dto.variables):

            for variable_id, variable_record in self.dto.variables.items():

                if variable_id not in processed_vars:

                    variable, warn = variable_record.to_variable(model=model,
                                                                 types=variables.get(variable_id, {'regulator'}),
                                                                 name=variable_record.name,
                                                                 aliases=variable_record.aliases,
                                                                 coefficients=variable_record.coefficients,
                                                                 active_coefficient=variable_record.active_coefficient)

                    if warn:
                        self.warnings.append(partial(sbml_warning, warn))

                    model.add(variable, 'regulator')

        return model

    def _reverse_f_id(self, variable):

        if variable.is_reaction():

            return f_id(variable.id, F_REACTION_REV)

        elif variable.is_metabolite():

            return f_id(variable.id, F_SPECIE_REV)

        elif variable.is_gene():

            return f_id(variable.id, F_GENE_REV)

        elif variable.is_interaction():

            if variable.target:

                target_id = self._reverse_f_id(variable.target)

                return f_id(target_id, F_TRANSITION_REV)

            else:

                return f_id(variable.id, F_TRANSITION_REV)

        else:

            return variable.id

    @staticmethod
    def _expression_replace(expression_string, replacements):

        for key, value in replacements.items():
            expression_string = expression_string.replace(key, value)

        return expression_string

    def write(self):

        if self.dto is None:
            raise OSError('SBML file is not open')

        if self.dto.doc is None:
            raise OSError('SBML file is not open')

        if self.dto.model is None:
            raise OSError(f'SBML file is not open')

        # -----------------------------------------------------------------------------
        # Compartments
        # -----------------------------------------------------------------------------
        default_compartment = None

        for compartment_id, compartment_name in self.model.compartments.items():
            sbml_compartment = self.dto.model.createCompartment()
            sbml_compartment.setId(compartment_id)
            sbml_compartment.setName(compartment_name)
            sbml_compartment.setConstant(True)

            if default_compartment is None:
                default_compartment = compartment_id

        if default_compartment is None:
            default_compartment = 'e'
            sbml_compartment = self.dto.model.createCompartment()
            sbml_compartment.setId('e')
            sbml_compartment.setName('extracellular')
            sbml_compartment.setConstant(True)

        # -----------------------------------------------------------------------------
        # Variables/Qualitative Species - Targets
        # -----------------------------------------------------------------------------
        processed_regulators = []

        for target in self.model.yield_targets():

            qual_species = self.dto.qual_plugin.createQualitativeSpecies()

            target_id = self._reverse_f_id(target)
            qual_species.setId(target_id)

            if hasattr(target, 'compartment'):

                if target.compartment is None:
                    compartment = default_compartment

                else:
                    compartment = target.compartment

            else:
                compartment = default_compartment

            qual_species.setCompartment(compartment)

            qual_species.setConstant(False)
            qual_species.setName(target.name)

            qual_species.setInitialLevel(int(ceil(target.coefficient.active_coefficient)))
            qual_species.setMaxLevel(int(ceil(target.coefficient.maximum_coefficient)))

            processed_regulators.append(target.id)

        # -----------------------------------------------------------------------------
        # Variables/Qualitative Species - Regulators
        # -----------------------------------------------------------------------------

        for regulator in self.model.yield_regulators():

            if regulator.id not in processed_regulators:

                qual_species = self.dto.qual_plugin.createQualitativeSpecies()

                regulator_id = self._reverse_f_id(regulator)
                qual_species.setId(regulator_id)

                if hasattr(regulator, 'compartment'):

                    if regulator.compartment is None:
                        compartment = default_compartment

                    else:
                        compartment = regulator.compartment

                else:
                    compartment = default_compartment

                qual_species.setCompartment(compartment)

                qual_species.setConstant(False)
                qual_species.setName(regulator.name)

                qual_species.setInitialLevel(int(ceil(regulator.coefficient.active_coefficient)))
                qual_species.setMaxLevel(int(ceil(regulator.coefficient.maximum_coefficient)))

        # -----------------------------------------------------------------------------
        # Transitions - Interactions
        # -----------------------------------------------------------------------------
        for interaction in self.model.yield_interactions():

            transition = self.dto.qual_plugin.createTransition()

            transition_id = self._reverse_f_id(interaction)

            transition.setId(transition_id)
            transition.setName(interaction.name)

            # -----------------------------------------------------------------------------
            # Output
            # -----------------------------------------------------------------------------

            if interaction.target:
                output = transition.createOutput()

                target_id = self._reverse_f_id(interaction.target)

                output.setId(f'{target_id}_out')
                output.setQualitativeSpecies(target_id)

            # -----------------------------------------------------------------------------
            # Inputs
            # -----------------------------------------------------------------------------

            for regulator in interaction.yield_regulators():
                regulator_id = self._reverse_f_id(regulator)

                input_id = f'{regulator_id}_input'
                reg_input = transition.createInput()
                reg_input.setId(input_id)
                reg_input.setQualitativeSpecies(regulator_id)

            # -----------------------------------------------------------------------------
            # Default Term
            # -----------------------------------------------------------------------------
            default_term = transition.createDefaultTerm()
            default_assigned = False

            for coefficient, expression in interaction.regulatory_events.items():

                # default term assignment
                if expression.is_none and not default_assigned:
                    default_term.setResultLevel(int(coefficient))

                    default_assigned = True

                    continue

                # -----------------------------------------------------------------------------
                # Function Terms
                # -----------------------------------------------------------------------------
                function_term = transition.createFunctionTerm()
                function_term.setResultLevel(int(coefficient))

                # operators to be replaced
                replacements = {'&': '&&',
                                '|': '||',
                                '~': '!',
                                '=': '==',
                                '<==': '<=',
                                '>==': '>='}

                # reverse ids for the variables that must be replaced
                symbols_reverse_ids = {variable.id: self._reverse_f_id(variable)
                                       for variable in expression.variables.values()}

                replacements.update(symbols_reverse_ids)

                expression_string = expression.to_string()

                expression_string = self._expression_replace(expression_string=expression_string,
                                                             replacements=replacements)

                if not expression_string:
                    self.warnings.append(partial(sbml_warning, f'Empty expression to be set as a function term in '
                                                               f'transition {transition_id}'))

                    continue

                warning = set_math(transition_id, expression_string, function_term)

                if warning:
                    self.warnings.append(partial(sbml_warning, warning))

        write_sbml_doc(self.io, self.dto.doc)

    def close(self):

        if hasattr(self.io, 'close'):
            self.io.close()

    def clean(self):
        self._dto = None


class MetabolicSBML(Engine):

    def __init__(self, io, config, model=None):
        super().__init__(io, config, model)

    @property
    def model_type(self):
        return 'metabolic'

    @property
    def model(self):

        if self._model is None:

            identifier = self.dto.id

            if not identifier:
                identifier = self.get_identifier()

            return MetabolicModel(identifier=identifier)

        return self._model

    def parse_notes(self, notes):

        notes_store = {}

        for match in pattern_notes.finditer(notes):

            try:
                key, value = match.group("content").split(":", 1)

            except ValueError:
                self.warnings.append(partial(sbml_warning,
                                             "Unexpected content format {}.".format(match.group("content"))))
                continue

            value = value.strip()

            if value:
                notes_store[key.strip()] = value

        return notes_store

    @staticmethod
    def parse_symbols(fbc_association):

        gene_id = f_id(fbc_association.getGeneProduct(), F_GENE)

        symbol = Symbol(gene_id)

        return symbol

    def parse_leaves(self, fbc_association):

        if fbc_association.isGeneProductRef():

            return self.parse_symbols(fbc_association)

        else:

            return

    def _parse_gpa(self, fbc_association):

        if fbc_association.isFbcOr():

            args = [self._parse_gpa(child)
                    for child in fbc_association.getListOfAssociations()]

            return Or(variables=args)

        elif fbc_association.isFbcAnd():

            args = [self._parse_gpa(child)
                    for child in fbc_association.getListOfAssociations()]

            return And(variables=args)

        else:

            return self.parse_leaves(fbc_association)

    def parse_gpa(self, fbc_association):

        """
        Reads and parses a node of type math ASTNode into a boolean algebraic expression.

        :param fbc_association: FBCAssociation
        :return: Expression
        """

        try:

            symbolic = self._parse_gpa(fbc_association)

        except SyntaxError:

            self.warnings.append(partial(expression_warning,
                                         f'{fbc_association} cannot be parsed. Assigning empty expression instead'))

            symbolic = NoneAtom()

        return symbolic

    def get_identifier(self):

        if os.path.exists(self.io):
            _, identifier = os.path.split(self.io)
            return os.path.splitext(identifier)[0]

        return 'model'

    def _open_to_read(self):

        self._dto = DataTransferObject()

        # -----------------------------------------------------------------------------
        # Doc, Model and FBC Plugin
        # -----------------------------------------------------------------------------

        self.dto.doc = get_sbml_doc_to_read(self.io)

        self.dto.model = self.dto.doc.getModel()

        if self.dto.model is None:
            raise OSError(f'{self.io} is not a valid input. Model SBML section is missing. '
                          f'Provide a correct path or file handler')

        identifier = self.dto.model.getIdAttribute()

        if not identifier:
            self.warnings.append(partial(sbml_warning, 'Model identifier is not encoded in the SBML file'))

            identifier = self.get_identifier()

        self.dto.id = identifier
        self.dto.name = self.dto.model.getName()

    def _open_to_write(self):

        self._dto = DataTransferObject()

        # -----------------------------------------------------------------------------
        # Doc, Model and FBC Plugin
        # -----------------------------------------------------------------------------

        self.dto.doc = get_sbml_doc_to_write(self.io,
                                             level=3,
                                             version=1,
                                             packages=('fbc',),
                                             packages_version=(2,),
                                             packages_required=(False,),
                                             sbo_term=True)

        if self.dto.model is None:

            self.dto.model = self.dto.doc.createModel()

        else:

            self.dto.model = self.dto.doc.getModel()

        # fbc plugin is added by get_sbml_doc_to_write
        self.dto.fbc_plugin = self.dto.model.getPlugin('fbc')
        self.dto.fbc_plugin.setStrict(True)

        if self.model.id is not None:
            self.dto.model.setId(self.model.id)
            self.dto.model.setMetaId('meta_' + self.model.id)

        else:
            self.dto.model.setMetaId('meta_model')

        if self.model.name is not None:
            self.dto.model.setName(self.model.name)

    def open(self, mode='r'):

        if mode == 'r':

            return self._open_to_read()

        elif mode == 'w':

            return self._open_to_write()

        else:
            raise ValueError(f'{mode} mode is not recognized. Try one of the following: r, w')

    def parse(self):

        if self.dto is None:
            raise OSError('SBML file is not open')

        if self.dto.id is None:
            raise OSError('SBML file is not open')

        if self.dto.doc is None:
            raise OSError('SBML file is not open')

        if self.dto.model is None:
            raise OSError(f'SBML file is not open')

        self.dto.fbc_plugin = self.dto.model.getPlugin("fbc")

        if not self.dto.fbc_plugin:
            self.warnings.append(partial(sbml_warning, "SBML model does not have fbc plugin"))

        else:
            if not self.dto.fbc_plugin.isSetStrict():
                self.warnings.append(partial(sbml_warning, 'SBML model fbc plugin is not set to strict. It must '
                                                           'fbc:strict="true"'))

            doc_fbc = self.dto.doc.getPlugin("fbc")
            fbc_version = doc_fbc.getPackageVersion()

            # fbc 1 to 2. If fails, an import error is launched
            if fbc_version == 1:
                self.warnings.append(partial(sbml_warning, 'Models should be encoded using fbc version 2. Converting '
                                                           'fbc v1 to fbc v2'))

                convert_fbc(self.dto.doc)

        # -----------------------------------------------------------------------------
        # Model id, name, etc
        # -----------------------------------------------------------------------------

        self.dto.level = self.dto.model.getLevel()
        self.dto.version = self.dto.model.getVersion()

        has_history = self.dto.model.isSetModelHistory()

        if has_history:

            history = self.dto.model.getModelHistory()

            created = None
            if history.isSetCreatedDate():
                created = history.getCreatedDate()

            creators = history.getListCreators()

            self.dto.history = History(created, creators)

        # -----------------------------------------------------------------------------
        # Compartments
        # -----------------------------------------------------------------------------

        for compartment in self.dto.model.getListOfCompartments():
            self.dto.compartments[compartment.getIdAttribute()] = CompartmentRecord(id=compartment.getIdAttribute(),
                                                                                    name=compartment.getName())

        # -----------------------------------------------------------------------------
        # Metabolites and Extracellular metabolites
        # -----------------------------------------------------------------------------

        metabolites = {}
        extracellular_metabolites = {}

        if self.dto.model.getNumSpecies() == 0:
            self.warnings.append(partial(sbml_warning, "SBML model does not have species/metabolites"))

        for met in self.dto.model.getListOfSpecies():

            met_id = f_id(met.getIdAttribute(), F_SPECIE)
            met_name = met.getName()
            met_aliases = {met_id, met_name}

            if not met_name:
                met_name = met_id

            met_notes = met.getNotesString()
            met_annotation = met.getAnnotationString()
            met_compartment = met.getCompartment()

            met_fbc = met.getPlugin("fbc")
            if met_fbc:
                met_charge = met_fbc.getCharge()
                met_formula = met_fbc.getChemicalFormula()
            else:
                if met.isSetCharge():
                    met_charge = met.getCharge()
                    met_formula = None
                else:
                    met_charge = None
                    met_formula = None

            met_record = VariableRecord(id=met_id,
                                        name=met_name,
                                        aliases=met_aliases,
                                        notes=met_notes,
                                        annotation=met_annotation,
                                        compartment=met_compartment,
                                        charge=met_charge,
                                        formula=met_formula)

            self.variables[met_id].add('metabolite')

            if met.getBoundaryCondition() is True:
                # extracellular metabolites
                extracellular_metabolites[met_id] = met_record

            metabolites[met_id] = met_record

            self.dto.variables[met_id] = met_record

        self.dto.metabolites = metabolites
        self.dto.extracellular_metabolites = extracellular_metabolites

        # -----------------------------------------------------------------------------
        # Genes
        # -----------------------------------------------------------------------------

        genes = {}

        # parsing genes encoded in the gene products section of the sbml
        if self.dto.fbc_plugin:

            if self.dto.fbc_plugin.getNumGeneProducts() == 0:
                self.warnings.append(partial(sbml_warning, "SBML model fbc plugin does not have gene products/genes"))

            for gene in self.dto.fbc_plugin.getListOfGeneProducts():

                gene_id = f_id(gene.getIdAttribute(), F_GENE)
                gene_name = gene.getName()

                if not gene_name:
                    gene_name = gene_id

                gene_aliases = {gene_id, gene_name}

                gene_notes = gene.getNotesString()
                gene_annotation = gene.getAnnotationString()

                gene_record = VariableRecord(id=gene_id,
                                             name=gene_name,
                                             aliases=gene_aliases,
                                             notes=gene_notes,
                                             annotation=gene_annotation)

                self.variables[gene_id].add('gene')

                genes[gene_id] = gene_record
                self.dto.variables[gene_id] = gene_record

        self.dto.genes = genes

        # -----------------------------------------------------------------------------
        # Reactions
        # -----------------------------------------------------------------------------

        reactions = {}

        if self.dto.model.getNumReactions() == 0:
            self.warnings.append(partial(sbml_warning, "SBML model does not have reactions"))

        for reaction in self.dto.model.getListOfReactions():

            rxn_id = f_id(reaction.getIdAttribute(), F_REACTION)
            rxn_name = reaction.getName()

            if not rxn_name:
                rxn_name = rxn_id

            rxn_aliases = {rxn_id, rxn_name}

            rxn_notes = self.parse_notes(reaction.getNotesString())
            rxn_annotation = reaction.getAnnotationString()

            # ------------------------------------------------
            # Reaction fbc plugin is of the upmost importance to parse reaction properties
            # ------------------------------------------------

            rxn_fbc = reaction.getPlugin("fbc")

            # ------------------------------------------------
            # Reaction Bounds
            # ------------------------------------------------

            if rxn_fbc:

                rxn_bounds = [ModelConstants.REACTION_LOWER_BOUND, ModelConstants.REACTION_UPPER_BOUND]

                # bounds in fbc parameters section
                bounds_ids = (rxn_fbc.getLowerFluxBound(), rxn_fbc.getUpperFluxBound())

                for i, bound_id in enumerate(bounds_ids):

                    if bound_id:

                        bound_parameter = self.dto.model.getParameter(bound_id)

                        if bound_parameter and bound_parameter.getConstant() and bound_parameter.getValue() is not None:

                            rxn_bounds[i] = bound_parameter.getValue()

                        else:

                            self.warnings.append(
                                partial(sbml_warning, f"Incorrect {bound_parameter} bound for {reaction} reaction. "
                                                      f"Set to default"))

                    else:

                        self.warnings.append(partial(sbml_warning, f"Bound for {reaction} reaction not found in the "
                                                                   f"SBML model fbc plugin. Set to "
                                                                   f"default. Try to set all bounds explicitly on all "
                                                                   f"reactions"))

            elif reaction.isSetKineticLaw():

                self.warnings.append(partial(sbml_warning, f"{reaction} reaction fbc plugin not found. This might "
                                                           f"hinder reaction parsing"))

                self.warnings.append(partial(
                    sbml_warning, f"Bounds have been detected in kinetic laws for {reaction} reaction. "
                                  f"Try to set all bounds explicitly on all reactions using the fbc plugin, as mewpy "
                                  f"can miss sometimes kinetic laws"))

                # bounds encoded in the kinetic law. Not advised
                kinetic_law = reaction.getKineticLaw()

                rxn_bounds = [ModelConstants.REACTION_LOWER_BOUND, ModelConstants.REACTION_UPPER_BOUND]

                # parameters of the kinetic law
                kinetic_parameters = ('LOWER_BOUND', 'UPPER_BOUND')

                for i, kinetic_parameter in enumerate(kinetic_parameters):

                    bound = kinetic_law.getParameter(kinetic_parameter)

                    if bound:
                        rxn_bounds[i] = bound.getValue()

                    else:
                        self.warnings.append(partial(sbml_warning,
                                                     f"{kinetic_parameter} has not been detected. "
                                                     f"{kinetic_parameter} has been set to default. "
                                                     f"Try to set all bounds explicitly on all reactions using "
                                                     f"the fbc plugin"))

            else:

                self.warnings.append(partial(sbml_warning, f"{reaction} reaction fbc plugin not found. This might "
                                                           f"hinder reaction parsing"))

                self.warnings.append(partial(sbml_warning,
                                             f"Bounds have not been detected. Bounds have been set to default. Try to "
                                             f"set all bounds explicitly on all reactions using the fbc plugin"))

                rxn_bounds = [ModelConstants.REACTION_LOWER_BOUND, ModelConstants.REACTION_UPPER_BOUND]

            # ------------------------------------------------
            # Reaction stoichiometry and metabolites
            # ------------------------------------------------

            rxn_stoichiometry = {}
            rxn_metabolites = {}
            rxn_products = {}
            rxn_reactants = {}

            # reactants
            for reactant in reaction.getListOfReactants():

                reactant_id = f_id(reactant.getSpecies(), F_SPECIE)

                if reactant_id not in self.dto.metabolites:
                    raise ValueError(f"{reactant_id} reactant of {reaction} reaction "
                                     f"is not listed as specie in the SBML Model.")

                reactant_record = self.dto.metabolites[reactant_id]

                rxn_metabolites[reactant_id] = reactant_record
                rxn_reactants[reactant_id] = reactant_record
                rxn_stoichiometry[reactant_id] = -reactant.getStoichiometry()

            # products
            for product in reaction.getListOfProducts():

                product_id = f_id(product.getSpecies(), F_SPECIE)

                if product_id not in self.dto.metabolites:
                    raise ValueError(f"{product_id} product of {reaction} reaction "
                                     f"is not listed as specie in the SBML Model.")

                product_record = self.dto.metabolites[product_id]

                rxn_metabolites[product_id] = product_record
                rxn_products[product_id] = product_record
                rxn_stoichiometry[product_id] = product.getStoichiometry()

            # ------------------------------------------------
            # Reaction Gene-Protein-Reaction rule
            # ------------------------------------------------

            symbolic = NoneAtom()

            if rxn_fbc:

                # if the gpr is encoded in the reaction fbc plugin.
                # It is advised.

                gpa = rxn_fbc.getGeneProductAssociation()

                if gpa is not None:
                    symbolic = self.parse_gpa(gpa.getAssociation())

            else:

                self.warnings.append(partial(sbml_warning,
                                             "Please use fbc plugin fbc:gpr to encode gprs in the future, "
                                             "as parsing gprs from notes might be troublesome"))

                # Else the gpr parsing tries to find the gpr rule (string) within the notes

                gpr_rule = rxn_notes.get('GENE ASSOCIATION',
                                         rxn_notes.get('GENE_ASSOCIATION', None))

                if gpr_rule is None:

                    self.warnings.append(partial(sbml_warning,
                                                 "GPR was not found within the reaction's notes section"))

                else:

                    gpr_rule = ' '.join(f_id(child, F_GENE) for child in gpr_rule.split(' '))

                    symbolic, warning = build_symbolic(expression=gpr_rule)

                    if warning:
                        self.warnings.append(partial(expression_warning, warning))

            rxn_genes = {}
            for symbol in symbolic.atoms(symbols_only=True):

                if rxn_fbc:

                    if symbol.name in self.dto.genes:

                        rxn_genes[symbol.name] = self.dto.genes[symbol.name]

                        continue

                    else:

                        self.warnings.append(partial(sbml_warning,
                                                     f'{symbol.name} is not listed in the SBML model fbc plugin'))

                gene_record = VariableRecord(id=symbol.name,
                                             name=symbol.name,
                                             aliases={symbol.name, symbol.value})

                self.variables[symbol.name].add('gene')

                self.dto.variables[symbol.name] = gene_record
                self.dto.genes[symbol.name] = gene_record

                rxn_genes[symbol.name] = gene_record

            # -----------------------------------------------------------------------------
            # GPR Function term
            # -----------------------------------------------------------------------------
            function_term = FunctionTerm(id='gpr_term', symbolic=symbolic, coefficient=1)

            # ------------------------------------------------
            # Building reaction record and assigning to containers
            # ------------------------------------------------

            reaction_record = VariableRecord(id=rxn_id,
                                             name=rxn_name,
                                             aliases=rxn_aliases,
                                             notes=rxn_notes,
                                             annotation=rxn_annotation,
                                             bounds=tuple(rxn_bounds),
                                             genes=rxn_genes,
                                             gpr=function_term,
                                             metabolites=rxn_metabolites,
                                             products=rxn_products,
                                             reactants=rxn_reactants,
                                             stoichiometry=rxn_stoichiometry)

            self.variables[rxn_id].add('reaction')

            reactions[rxn_id] = reaction_record

            self.dto.variables[rxn_id] = reaction_record

        self.dto.reactions = reactions

        # Some transport reactions have fictitious extracellular metabolites, encoded with the prefix _b,
        # such as M_a_b. In this case, a exchange reaction is created for each metabolite

        extracellular_reactions = {}

        for extracellular_met in self.dto.extracellular_metabolites.values():
            # ------------------------------------------------
            # Building reaction record and assigning to containers
            # ------------------------------------------------

            reaction_record = VariableRecord(id=f'EX_{extracellular_met.id}',
                                             name=f'EX_{extracellular_met.id}',
                                             bounds=(ModelConstants.REACTION_LOWER_BOUND,
                                                     ModelConstants.REACTION_UPPER_BOUND),
                                             metabolites={extracellular_met.id: extracellular_met},
                                             reactants={extracellular_met.id: extracellular_met},
                                             stoichiometry={extracellular_met.id: -1})

            self.variables[f'EX_{extracellular_met.id}'].add('reaction')

            extracellular_reactions[f'EX_{extracellular_met.id}'] = reaction_record

            self.dto.variables[f'EX_{extracellular_met.id}'] = reaction_record

            self.warnings.append(partial(sbml_warning,
                                         f'EX_{extracellular_met.id} reaction added '
                                         f'for metabolite {extracellular_met.id}'))

        self.dto.extracellular_reactions = extracellular_reactions
        self.dto.reactions.update(extracellular_reactions)

        # -----------------------------------------------------------------------------
        # Objective
        # -----------------------------------------------------------------------------

        model_objective = {}

        if self.dto.fbc_plugin:

            objectives = self.dto.fbc_plugin.getListOfObjectives()

            if objectives is None:
                self.warnings.append(partial(sbml_warning, "listOfObjectives element not found"))

            elif objectives.size() == 0:
                self.warnings.append(partial(sbml_warning, "No objective in listOfObjectives"))

            elif not objectives.getActiveObjective():
                self.warnings.append(partial(sbml_warning, "No active objective in listOfObjectives"))

            else:
                active_objectives = objectives.getActiveObjective()
                objective = self.dto.fbc_plugin.getObjective(active_objectives)

                direction = objective.getType()

                for flux_objective in objective.getListOfFluxObjectives():

                    flux_objective_id = f_id(flux_objective.getReaction(), F_REACTION)

                    objective_rxn = self.dto.reactions.get(flux_objective_id, None)

                    if objective_rxn is None:
                        self.warnings.append(partial(sbml_warning,
                                                     f"Objective {flux_objective_id} reaction "
                                                     f"not found in the SBML model"))

                        continue

                    coef = flux_objective.getCoefficient()

                    if direction == 'minimize':
                        coef = -coef

                    model_objective[flux_objective_id] = coef

        else:
            self.warnings.append(partial(sbml_warning,
                                         f"Objective might be encoded in kinetic laws of a given reaction. However, "
                                         f"mewpy does not handle kinetic laws. The objective has not been set. Try "
                                         f"to set the objective explicitly on the fbc plugin"))

        if len(model_objective) == 0:
            self.warnings.append(partial(sbml_warning, "No objective found for the model"))

        self.dto.objective = model_objective

    def read(self,
             model: Union['Model', 'MetabolicModel', 'RegulatoryModel'] = None,
             variables=None):

        if not model:
            model: Union['Model', 'MetabolicModel', 'RegulatoryModel'] = self.model

        if not variables:
            variables = self.variables

        if self.dto.id:
            model._id = self.dto.id

        if self.dto.name:
            model.name = self.dto.name

        model.compartments = {compartment.id: compartment.name
                              for compartment in self.dto.compartments.values()}

        processed_metabolites = set()
        processed_genes = set()

        for rxn_id, rxn_record in self.dto.reactions.items():

            genes = {}

            for gene_id, gene_record in rxn_record.genes.items():

                gene, warning = gene_record.to_variable(model=model,
                                                        types=variables.get(gene_id, {'gene'}),
                                                        name=gene_record.name,
                                                        aliases=gene_record.aliases)

                if warning:
                    self.warnings.append(partial(sbml_warning, warning))

                genes[gene_id] = gene

                processed_genes.add(gene_id)

            stoichiometry = {}

            for met_id, met_record in rxn_record.metabolites.items():

                met, warning = met_record.to_variable(model=model,
                                                      types=variables.get(met_id, {'metabolite'}),
                                                      name=met_record.name,
                                                      aliases=met_record.aliases,
                                                      compartment=met_record.compartment,
                                                      charge=met_record.charge,
                                                      formula=met_record.formula)

                if warning:
                    self.warnings.append(partial(sbml_warning, warning))

                coef = rxn_record.stoichiometry[met_id]

                stoichiometry[met] = coef

                processed_metabolites.add(met_id)

            gpr = Expression(symbolic=rxn_record.gpr.symbolic, variables=genes)

            rxn, warning = rxn_record.to_variable(model=model,
                                                  types=variables.get(rxn_id, {'reaction'}),
                                                  bounds=rxn_record.bounds,
                                                  gpr=gpr,
                                                  stoichiometry=stoichiometry)

            if warning:
                self.warnings.append(partial(sbml_warning, warning))

            model.add(rxn, 'reaction')

        to_append = []

        for met_id, met_record in self.dto.metabolites.items():

            if met_id not in processed_metabolites:
                met, warning = met_record.to_variable(model=model,
                                                      types=variables.get(met_id, {'metabolite'}),
                                                      name=met_record.name,
                                                      aliases=met_record.aliases,
                                                      compartment=met_record.compartment,
                                                      charge=met_record.charge,
                                                      formula=met_record.formula)

                if warning:
                    self.warnings.append(partial(sbml_warning, warning))

                to_append.append(met)

        for gene_id, gene_record in self.dto.genes.items():

            if gene_id not in processed_genes:
                gene, warning = gene_record.to_variable(model=model,
                                                        types=variables.get(gene_id, {'gene'}),
                                                        name=gene_record.name,
                                                        aliases=gene_record.aliases)

                if warning:
                    self.warnings.append(partial(sbml_warning, warning))

                to_append.append(gene)

        model.add(to_append)

        model.objective = self.dto.objective

        return model

    def write(self):

        if self.dto is None:
            raise OSError('SBML file is not open')

        if self.dto.doc is None:
            raise OSError('SBML file is not open')

        if self.dto.model is None:
            raise OSError(f'SBML file is not open')

        # -----------------------------------------------------------------------------
        # Units
        # -----------------------------------------------------------------------------
        units = self.config.get('units', False)
        unit_definition = None
        if units:

            unit_definition = self.dto.model.createUnitDefinition()
            unit_definition.setId(UNIT_ID)

            for unit in UNITS:
                _unit = unit_definition.createUnit()
                _unit.setKind(unit.kind)
                _unit.setExponent(unit.exponent)
                _unit.setScale(unit.scale)
                _unit.setMultiplier(unit.multiplier)

        # -----------------------------------------------------------------------------
        # Constants and parameters
        # -----------------------------------------------------------------------------
        add_sbml_parameter(sbml_model=self.dto.model,
                           parameter_id=LOWER_BOUND_ID,
                           value=ModelConstants.REACTION_LOWER_BOUND,
                           constant=True,
                           sbo=SBO_DEFAULT_FLUX_BOUND)

        add_sbml_parameter(sbml_model=self.dto.model,
                           parameter_id=UPPER_BOUND_ID,
                           value=ModelConstants.REACTION_UPPER_BOUND,
                           constant=True,
                           sbo=SBO_DEFAULT_FLUX_BOUND)

        add_sbml_parameter(sbml_model=self.dto.model,
                           parameter_id=ZERO_BOUND_ID,
                           value=0,
                           constant=True,
                           sbo=SBO_DEFAULT_FLUX_BOUND)

        add_sbml_parameter(sbml_model=self.dto.model,
                           parameter_id=BOUND_MINUS_INF,
                           value=-float("Inf"),
                           constant=True,
                           sbo=SBO_DEFAULT_FLUX_BOUND)

        add_sbml_parameter(sbml_model=self.dto.model,
                           parameter_id=BOUND_PLUS_INF,
                           value=float("Inf"),
                           constant=True,
                           sbo=SBO_DEFAULT_FLUX_BOUND)

        # -----------------------------------------------------------------------------
        # Compartments
        # -----------------------------------------------------------------------------
        for compartment_id, compartment_name in self.model.compartments.items():
            sbml_compartment = self.dto.model.createCompartment()
            sbml_compartment.setId(compartment_id)
            sbml_compartment.setName(compartment_name)
            sbml_compartment.setConstant(True)

        # -----------------------------------------------------------------------------
        # Metabolites
        # -----------------------------------------------------------------------------
        for metabolite in self.model.yield_metabolites():
            sbml_specie = self.dto.model.createSpecies()

            sbml_specie.setId(f_id(metabolite.id, F_SPECIE_REV))
            sbml_specie.setConstant(False)
            sbml_specie.setBoundaryCondition(False)
            sbml_specie.setHasOnlySubstanceUnits(False)
            sbml_specie.setName(metabolite.name)
            sbml_specie.setCompartment(metabolite.compartment)
            specie_fbc = sbml_specie.getPlugin('fbc')
            specie_fbc.setCharge(metabolite.charge)
            specie_fbc.setChemicalFormula(metabolite.formula)

        # -----------------------------------------------------------------------------
        # Genes
        # -----------------------------------------------------------------------------
        for gene in self.model.yield_genes():
            gene_product = self.dto.fbc_plugin.createGeneProduct()

            gene_product.setId(f_id(gene.id, F_GENE_REV))
            gene_product.setName(gene.name)
            gene_product.setLabel(gene.id)

        # -----------------------------------------------------------------------------
        # Objective
        # -----------------------------------------------------------------------------
        objective = self.dto.fbc_plugin.createObjective()
        objective.setId('obj')
        objective.setType('maximize')
        self.dto.fbc_plugin.setActiveObjectiveId('obj')

        for reaction, coefficient in self.model.objective.items():
            flux_objective = objective.createFluxObjective()
            flux_objective.setReaction(f_id(reaction.id, F_REACTION_REV))
            flux_objective.setCoefficient(coefficient)

        # -----------------------------------------------------------------------------
        # Reactions
        # -----------------------------------------------------------------------------
        for reaction in self.model.yield_reactions():

            sbml_reaction = self.dto.model.createReaction()
            sbml_reaction.setId(f_id(reaction.id, F_REACTION_REV))
            sbml_reaction.setName(reaction.name)
            sbml_reaction.setFast(False)
            sbml_reaction.setReversible(reaction.reversibility)

            # -----------------------------------------------------------------------------
            # Stoichiometry
            # -----------------------------------------------------------------------------
            for metabolite, st in reaction.stoichiometry.items():

                if st < 0:
                    sbml_reactant = sbml_reaction.createReactant()
                    sbml_reactant.setSpecies(f_id(metabolite.id, F_SPECIE_REV))
                    sbml_reactant.setStoichiometry(-st)
                    sbml_reactant.setConstant(True)

                else:
                    sbml_product = sbml_reaction.createProduct()
                    sbml_product.setSpecies(f_id(metabolite.id, F_SPECIE_REV))
                    sbml_product.setStoichiometry(st)
                    sbml_product.setConstant(True)

            # -----------------------------------------------------------------------------
            # Bounds
            # -----------------------------------------------------------------------------
            sbml_rxn_fbc = sbml_reaction.getPlugin('fbc')

            lb_parameter_id = get_sbml_lb_id(sbml_model=self.dto.model,
                                             reaction=reaction,
                                             unit_definition=unit_definition)

            if lb_parameter_id is None:
                lb_parameter_id = f'{f_id(reaction.id, F_REACTION_REV)}_lb'

                add_sbml_parameter(sbml_model=self.dto.model,
                                   parameter_id=lb_parameter_id,
                                   value=reaction.lower_bound,
                                   sbo=SBO_FLUX_BOUND,
                                   constant=True,
                                   unit_definition=unit_definition)

            sbml_rxn_fbc.setLowerFluxBound(lb_parameter_id)

            ub_parameter_id = get_sbml_ub_id(sbml_model=self.dto.model,
                                             reaction=reaction,
                                             unit_definition=unit_definition)

            if ub_parameter_id is None:
                ub_parameter_id = f'{f_id(reaction.id, F_REACTION_REV)}_lb'

                add_sbml_parameter(sbml_model=self.dto.model,
                                   parameter_id=ub_parameter_id,
                                   value=reaction.upper_bound,
                                   sbo=SBO_FLUX_BOUND,
                                   constant=True,
                                   unit_definition=unit_definition)

            sbml_rxn_fbc.setUpperFluxBound(ub_parameter_id)

            # -----------------------------------------------------------------------------
            # Reaction Gene-Protein-Reaction rule
            # -----------------------------------------------------------------------------
            set_gpr(self, sbml_warning, reaction, sbml_rxn_fbc)

        write_sbml_doc(self.io, self.dto.doc)

    def close(self):

        if hasattr(self.io, 'close'):
            self.io.close()

    def clean(self):
        self._dto = None



from mewpy.mew.lp.linear_problem import LinearProblem
from mewpy.mew.lp.linearizer import MetabolicLinearizer, GPRLinearizer, LogicLinearizer, InteractionLinearizer
from mewpy.mew.lp.notification import Notification
from mewpy.mew.lp.linear_containers import VariableContainer, ConstraintContainer
